# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (
    Any, Callable, Iterator, List, Optional, Sequence, Tuple, Union)

import numpy
from numpy.lib.recfunctions import merge_arrays
from numpy.typing import NDArray
from typing_extensions import TypeAlias, TypeGuard

from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    ConnectionsArray)

_ItemType: TypeAlias = numpy.floating
_Items: TypeAlias = Union[Tuple[NDArray[_ItemType], ...], NDArray[_ItemType],
                          Tuple[List[numpy.floating], ...]]


def _is_listable(value: Any) -> TypeGuard[Sequence[Any]]:
    return hasattr(value, "__len__")


class ConnectionHolder(object):
    """
    Holds a set of connections to be returned in a PyNN-specific format.
    """

    __slots__ = (
        # A list of items of data that are to be present in each element
        "__data_items_to_return",

        # True if the values should be returned as a list of tuples,
        # False if they should be returned as a tuple of matrices
        "__as_list",

        # The number of atoms in the pre-vertex
        "__n_pre_atoms",

        # The number of atoms in the post-vertex
        "__n_post_atoms",

        # A list of the connections that have been added
        "__connections",

        # The merged connections formed just before the data is read
        "__data_items",

        # Additional fixed values to be added to the data returned,
        # with the same values per synapse, as a list of tuples of
        # (field name, value)
        "__fixed_values",

        # A callback to call with the data when finished
        "__notify"
    )

    def __init__(
            self, data_items_to_return: Optional[Sequence[str]], as_list: bool,
            n_pre_atoms: int, n_post_atoms: int,
            connections: Optional[List[ConnectionsArray]] = None,
            fixed_values: Optional[List[Tuple[str, int]]] = None,
            notify: Optional[Callable[['ConnectionHolder'], None]] = None):
        """
        :param data_items_to_return: A list of data fields to be returned
        :param as_list:
            True if the data will be returned as a list, False if it is to be
            returned as a matrix (or series of matrices)
        :param n_pre_atoms: The number of atoms in the pre-vertex
        :param n_post_atoms: The number of atoms in the post-vertex
        :param connections:
            Any initial connections, as a numpy structured array of
            source, target, weight and delay
        :param fixed_values:
            A list of tuples of field names and fixed values to be appended
            to the other fields per connection, formatted as
            `[(field_name, value), ...]`.

            .. note::
                If the field is to be returned, the name must also
                appear in data_items_to_return, which determines the order of
                items in the result.
        :param notify:
            A callback to call when the connections have all been added.
            This should accept a single parameter, which will contain the
            data requested
        """
        self.__data_items_to_return = data_items_to_return
        self.__as_list = as_list
        self.__n_pre_atoms = n_pre_atoms
        self.__n_post_atoms = n_post_atoms
        self.__connections: Optional[List[NDArray]] = connections
        self.__data_items: Optional[_Items] = None
        self.__notify = notify
        self.__fixed_values = fixed_values

    def add_connections(self, connections: ConnectionsArray) -> None:
        """
        Add connections to the holder to be returned.

        :param connections:
            The connection to add, as a numpy structured array of
            source, target, weight and delay
        """
        if self.__connections is None:
            self.__connections = list()
        self.__connections.append(connections)

    @property
    def connections(self) -> List[ConnectionsArray]:
        """
        The connections stored.
        """
        return self.__connections or []

    def finish(self) -> None:
        """
        Finish adding connections.
        """
        if self.__notify is not None:
            self.__notify(self)

    def _get_data_items(self) -> _Items:
        """
        Merges the connections into the result data format.
        """
        # If there are already merged connections cached, return those
        if self.__data_items is not None:
            return self.__data_items

        if not self.__connections:
            # If there are no connections added, raise an exception
            if self.__connections is None:
                raise NotImplementedError(
                    f"Connections are only set after run has been called, "
                    f"even if you are trying to see the data before changes "
                    f"have been made. Try examining the "
                    f"{self.__data_items_to_return} after the call to run.")
            # If the list is empty assume on a virtual machine
            # with generation on machine
            if len(self.__connections) == 0:
                raise NotImplementedError(
                    "Connections list is empty. "
                    "This may be because you are using a virtual machine. "
                    "This projection creates connections on machine.")

        # Join all the connections that have been added (probably over multiple
        # sub-vertices of a population)
        connections: ConnectionsArray = numpy.concatenate(self.__connections)

        # If there are additional fixed values, merge them in
        if self.__fixed_values:
            # Generate a numpy type for the fixed values
            fixed_dtypes = [
                (f'{field[0]}', None)
                for field in self.__fixed_values]

            # Get the actual data as a record array
            fixed_data = numpy.asarray(
                tuple([field[1] for field in self.__fixed_values]),
                dtype=fixed_dtypes)

            # Tile the array to be the correct size
            fixed_values = numpy.tile(fixed_data, [len(connections), 1])

            # Add the fixed values to the connections
            connections = merge_arrays(
                (connections, fixed_values), flatten=True)

        # If we are returning a list...
        if self.__as_list:
            # ...sort by source then target
            order = numpy.lexsort(
                (connections["target"], connections["source"]))

            # There are no specific items to return, so just get
            # all the data
            if (self.__data_items_to_return is None or
                    not self.__data_items_to_return):
                data_items = connections[order]
            # There is more than one item to return, so let numpy do its magic
            elif len(self.__data_items_to_return) > 1:
                data_items = \
                    connections[order][self.__data_items_to_return]
            # There is 1 item to return, so make sure only one item exists
            else:
                data_items = \
                    connections[order][self.__data_items_to_return[0]]

            # Return in a format which can be understood by a FromListConnector
            items: List[List[numpy.floating]] = []
            # NB: The types in here are all wrong, but that's
            for data_item in data_items:
                if _is_listable(data_item):
                    items.append(list(data_item))
                else:
                    items.append(data_item)
            self.__data_items = tuple(items)

        else:
            if self.__data_items_to_return is None:
                return ()

            # Keep track of the matrices
            merged: List[NDArray[_ItemType]] = []
            for item in self.__data_items_to_return:
                # Build an empty matrix and fill it with NAN
                matrix = numpy.empty((self.__n_pre_atoms, self.__n_post_atoms))
                matrix.fill(numpy.nan)

                # Fill in the values that have data
                # TODO: Change this to sum the items with the same
                #       (source, target) pairs
                matrix[connections["source"], connections["target"]] = \
                    connections[item]

                # Store the matrix generated
                merged.append(matrix)

            # If there is only one matrix, use it directly
            # Otherwise use a tuple of the matrices
            self.__data_items = (
                merged[0] if len(merged) == 1 else tuple(merged))

        return self.__data_items

    def __getitem__(self, s: int) -> Union[
            numpy.floating, NDArray[numpy.floating],
            List[numpy.floating]]:
        data = self._get_data_items()
        return data[s]

    def __len__(self) -> int:
        data = self._get_data_items()
        return len(data)

    def __iter__(self) -> Iterator:
        data = self._get_data_items()
        return iter(data)

    def __str__(self) -> str:
        data = self._get_data_items()
        return data.__str__()

    def __repr__(self) -> str:
        data = self._get_data_items()
        return data.__repr__()

    def __getattr__(self, name: str) -> Any:
        """
        Gets the attributes from the NDArray if applicable

        :raises AttributeError: If the data does not have this attribute.
        """
        data = self._get_data_items()
        return getattr(data, name)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase


class AbstractHasAPlusAMinus(object, metaclass=AbstractBase):
    """
    An object that has :math:`A^+` and :math:`A^-` properties.
    """
    __slots__ = (
        '__a_plus',
        '__a_minus')

    def __init__(self) -> None:
        self.__a_plus: float = 0.0
        self.__a_minus: float = 0.0

    def set_a_plus_a_minus(self, a_plus: float, a_minus: float) -> None:
        """
        Set the values of :math:`A^+` and :math:`A^-`.

        :param a_plus: :math:`A^+`
        :param a_minus: :math:`A^-`
        """
        self.__a_plus = a_plus
        self.__a_minus = a_minus

    @property
    def A_plus(self) -> float:
        """
        Settable model parameter: :math:`A^+`
        """
        # pylint: disable=invalid-name
        return self.__a_plus

    @A_plus.setter
    def A_plus(self, new_value: float) -> None:
        # pylint: disable=invalid-name
        self.__a_plus = new_value

    @property
    def A_minus(self) -> float:
        """
        Settable model parameter: :math:`A^-`
        """
        # pylint: disable=invalid-name
        return self.__a_minus

    @A_minus.setter
    def A_minus(self, new_value: float) -> None:
        # pylint: disable=invalid-name
        self.__a_minus = new_value



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from enum import IntEnum
from typing import Optional, Sequence, TYPE_CHECKING

from spinn_utilities.overrides import overrides

from spinnman.model.enums import ExecutableType

from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import AbstractSDRAM

from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary, AbstractGeneratesDataSpecification)
from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl, ProvenanceWriter)
from spinn_front_end_common.interface.simulation import simulation_utilities
from spinn_front_end_common.utilities.constants import SIMULATION_N_BYTES

from spynnaker.pyNN.data import SpynnakerDataView

from .delay_extension_vertex import DelayExtensionVertex

if TYPE_CHECKING:
    from pacman.model.placements import Placement
    from spinn_front_end_common.interface.ds import DataSpecificationGenerator


class DelayExtensionMachineVertex(
        MachineVertex, ProvidesProvenanceDataFromMachineImpl,
        AbstractHasAssociatedBinary, AbstractGeneratesDataSpecification):
    """
    Vertex that implements a delay extension.
    Not expected to be directly referenced in user code.
    """

    __slots__ = (
        "__sdram",
        "__drop_late_spikes")

    class _DelayExtensionRegions(IntEnum):
        """
        Region indices.
        """
        SYSTEM = 0
        DELAY_PARAMS = 1
        PROVENANCE_REGION = 2
        TDMA_REGION = 3

    class _ExtraProvenanceDataEntries(IntEnum):
        """
        Indices into raw provenance data about delay extension vertices.
        """
        #: The number of packets received
        N_PACKETS_RECEIVED = 0
        #: The number of packets processed
        N_PACKETS_PROCESSED = 1
        #: The number of packets added to a send queue
        N_PACKETS_ADDED = 2
        #: The number of packets sent
        N_PACKETS_SENT = 3
        #: The number of times a buffer overflowed
        N_BUFFER_OVERFLOWS = 4
        #: The number of delays
        N_DELAYS = 5
        #: The number of packets lost due to saturation
        N_PACKETS_LOST_DUE_TO_COUNT_SATURATION = 6
        #: The number of packets with bad neuron IDs
        N_PACKETS_WITH_INVALID_NEURON_IDS = 7
        #: The number of packets lost due to an invalid key
        N_PACKETS_DROPPED_DUE_TO_INVALID_KEY = 8
        #: The number of packets that arrived too late
        N_LATE_SPIKES = 9
        #: The maximum number of packets queued for background processing
        MAX_BACKGROUND_QUEUED = 10
        #: The number of times the background queue overflowed
        N_BACKGROUND_OVERLOADS = 11

    N_EXTRA_PROVENANCE_DATA_ENTRIES = len(_ExtraProvenanceDataEntries)

    COUNT_SATURATION_NAME = "saturation_count"
    INVALID_NEURON_ID_COUNT_NAME = "invalid_neuron_count"
    INVALID_KEY_COUNT_NAME = "invalid_key_count"
    N_PACKETS_RECEIVED_NAME = "Number_of_packets_received"
    N_PACKETS_PROCESSED_NAME = "Number_of_packets_processed"
    MISMATCH_ADDED_FROM_PROCESSED_NAME = (
        "Number_of_packets_added_to_delay_slot")
    N_PACKETS_SENT_NAME = "Number_of_packets_sent"
    INPUT_BUFFER_LOST_NAME = "Times_the_input_buffer_lost_packets"
    N_LATE_SPIKES_NAME = "Number_of_late_spikes"
    DELAYED_FOR_TRAFFIC_NAME = "Number_of_times_delayed_to_spread_traffic"
    BACKGROUND_OVERLOADS_NAME = "Times_the_background_queue_overloaded"
    BACKGROUND_MAX_QUEUED_NAME = "Max_backgrounds_queued"

    def __init__(self, sdram: AbstractSDRAM, label: str, vertex_slice: Slice,
                 app_vertex: Optional[ApplicationVertex] = None):
        """
        :param sdram: The SDRAM required by the vertex
        :param label: The name of the vertex
        :param vertex_slice: The slice of the vertex
        :param app_vertex:
            The application vertex that caused this machine vertex to be
            created. If `None`, there is no such application vertex.
        """
        super().__init__(
            label, app_vertex=app_vertex, vertex_slice=vertex_slice)
        self.__sdram = sdram

    @property
    @overrides(MachineVertex.app_vertex)
    def app_vertex(self) -> DelayExtensionVertex:
        assert isinstance(self._app_vertex, DelayExtensionVertex)
        return self._app_vertex

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self) -> int:
        return self._DelayExtensionRegions.PROVENANCE_REGION

    @property
    @overrides(
        ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self) -> int:
        return self.N_EXTRA_PROVENANCE_DATA_ENTRIES

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self) -> AbstractSDRAM:
        return self.__sdram

    @overrides(ProvidesProvenanceDataFromMachineImpl.
               parse_extra_provenance_items)
    def parse_extra_provenance_items(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        (n_received, n_processed, n_added, n_sent, n_overflows, n_delays,
         n_sat, n_bad_neuron, n_bad_keys, n_late_spikes, max_bg,
         n_bg_overloads) = provenance_data

        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, self.COUNT_SATURATION_NAME, n_sat)
            if n_sat != 0:
                db.insert_report(
                    f"The delay extension {label} has dropped {n_sat} packets "
                    f"because during certain time steps a neuron was asked to "
                    f"spike more than 256 times. This causes a saturation on "
                    f"the count tracker which is a uint8. "
                    f"Reduce the packet rates, or modify the delay extension "
                    f"to have larger counters.")

            db.insert_core(
                x, y, p, self.INVALID_NEURON_ID_COUNT_NAME,
                n_bad_neuron)
            if n_bad_neuron != 0:
                db.insert_report(
                    f"The delay extension {label} has dropped {n_bad_neuron} "
                    f"packets because their neuron id was not valid. "
                    f"This is likely a routing issue. "
                    f"Please fix and try again")

            db.insert_core(
                x, y, p, self.INVALID_KEY_COUNT_NAME, n_bad_keys)
            if n_bad_keys != 0:
                db.insert_report(
                    f"The delay extension {label} has dropped {n_bad_keys} "
                    f"packets due to the packet key being invalid. "
                    f"This is likely a routing issue. "
                    f"Please fix and try again")

            db.insert_core(x, y, p, self.N_PACKETS_RECEIVED_NAME, n_received)

            db.insert_core(x, y, p, self.N_PACKETS_PROCESSED_NAME,
                           n_processed)
            if n_received != n_processed:
                db.insert_report(
                    f"The delay extension {label} only processed "
                    f"{n_processed} of {n_received} received packets.  "
                    f"This could indicate a fault.")

            db.insert_core(
                x, y, p, self.MISMATCH_ADDED_FROM_PROCESSED_NAME, n_added)
            if n_added != n_processed:
                db.insert_report(
                    f"The delay extension {label} only added {n_added} of "
                    f"{n_processed} processed packets.  This could indicate "
                    f"a routing or filtering fault")

            db.insert_core(x, y, p, self.N_PACKETS_SENT_NAME, n_sent)

            db.insert_core(
                x, y, p, self.INPUT_BUFFER_LOST_NAME, n_overflows)
            if n_overflows > 0:
                db.insert_report(
                    f"The input buffer for {label} lost packets on "
                    f"{n_overflows} occasions. This is often a sign that the "
                    f"system is running too quickly for the number of "
                    f"neurons per core.  "
                    f"Please increase the timer_tic or time_scale_factor "
                    f"or decrease the number of neurons per core.")

            db.insert_core(x, y, p, self.DELAYED_FOR_TRAFFIC_NAME, n_delays)

            db.insert_core(x, y, p, self.N_LATE_SPIKES_NAME, n_late_spikes)
            if n_late_spikes == 0:
                pass
            elif self.app_vertex.drop_late_spikes:
                db.insert_report(
                    f"On {label}, {n_late_spikes} packets were dropped from "
                    f"the input buffer, because they arrived too late to be "
                    f"processed in a given time step. "
                    "Try increasing the time_scale_factor located within the "
                    ".spynnaker.cfg file or in the pynn.setup() method.")
            else:
                db.insert_report(
                    f"On {label}, {n_late_spikes} packets arrived too late to "
                    f"be processed in a given time step. Try increasing the "
                    "time_scale_factor located within the .spynnaker.cfg file "
                    "or in the pynn.setup() method.")

            db.insert_core(
                x, y, p, self.BACKGROUND_MAX_QUEUED_NAME, max_bg)
            if max_bg > 1:
                db.insert_report(
                    f"On {label}, a maximum of {max_bg} background tasks "
                    f"were queued. Try increasing the time_scale_factor "
                    f"located within the .spynnaker.cfg file or in the "
                    f"pynn.setup() method.")

            db.insert_core(
                x, y, p, self.BACKGROUND_OVERLOADS_NAME, n_bg_overloads)
            if n_bg_overloads > 0:
                db.insert_report(
                    f"On {label}, the background queue overloaded "
                    f"{n_bg_overloads} times. "
                    f"Try increasing the time_scale_factor located within the "
                    ".spynnaker.cfg file or in the pynn.setup() method.")

    @overrides(MachineVertex.get_n_keys_for_partition)
    def get_n_keys_for_partition(self, partition_id: str) -> int:
        n_keys = super().get_n_keys_for_partition(partition_id)
        v = self.app_vertex
        n_colours = 2 ** v.n_colour_bits
        return n_keys * v.n_delay_stages * n_colours

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self) -> str:
        return "delay_extension.aplx"

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self) -> ExecutableType:
        return ExecutableType.USES_SIMULATION_INTERFACE

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec: DataSpecificationGenerator,
                                    placement: Placement) -> None:
        vertex = placement.vertex

        # Reserve memory:
        spec.comment("\nReserving memory space for data regions:\n\n")

        # ###################################################################
        # Reserve SDRAM space for memory areas:
        delay_params_sz = self.app_vertex.delay_params_size()

        spec.reserve_memory_region(
            region=self._DelayExtensionRegions.SYSTEM,
            size=SIMULATION_N_BYTES, label='setup')

        spec.reserve_memory_region(
            region=self._DelayExtensionRegions.DELAY_PARAMS,
            size=delay_params_sz, label='delay_params')

        # reserve region for provenance
        self.reserve_provenance_data_region(spec)

        assert isinstance(vertex, AbstractHasAssociatedBinary)
        self._write_setup_info(spec, vertex.get_binary_file_name())

        spec.comment("\n*** Spec for Delay Extension Instance ***\n\n")

        routing_infos = SpynnakerDataView.get_routing_infos()
        key = routing_infos.get_single_key_from(vertex)

        srcs = self.app_vertex.source_vertex.splitter.get_out_going_vertices(
            self.app_vertex.partition.identifier)
        for source_vertex in srcs:
            if source_vertex.vertex_slice == self.vertex_slice:
                r_info = routing_infos.get_info_from(
                    source_vertex, self.app_vertex.partition.identifier)
                incoming_key = r_info.key
                incoming_mask = r_info.mask
                break

        self.write_delay_parameters(
            spec, self.vertex_slice, key, incoming_key, incoming_mask)

        # End-of-Spec:
        spec.end_specification()

    def _write_setup_info(
            self, spec: DataSpecificationGenerator, binary_name: str) -> None:
        """
        :param spec:
        :param binary_name: the binary name
        """
        # Write this to the system region (to be picked up by the simulation):
        spec.switch_write_focus(self._DelayExtensionRegions.SYSTEM)
        spec.write_array(simulation_utilities.get_simulation_header_array(
            binary_name))

    def write_delay_parameters(
            self, spec: DataSpecificationGenerator, vertex_slice: Slice,
            key: Optional[int], incoming_key: int, incoming_mask: int) -> None:
        """
        Generate Delay Parameter data.

        :param spec:
        :param vertex_slice:
        :param key:
        :param incoming_key:
        :param incoming_mask:
        """
        # Write spec with commands to construct required delay region:
        spec.comment(
            f"Writing Delay Parameters for {vertex_slice.n_atoms} Neurons:\n")

        # Set the focus to the memory region 2 (delay parameters):
        spec.switch_write_focus(self._DelayExtensionRegions.DELAY_PARAMS)

        # Write header info to the memory region:
        # Write Key info for this core and the incoming key and mask:
        if key is None:
            spec.write_value(0)
            spec.write_value(0)
        else:
            spec.write_value(1)
            spec.write_value(data=key)
        spec.write_value(data=incoming_key)
        spec.write_value(data=incoming_mask)

        # Write the number of neurons in the block:
        spec.write_value(data=vertex_slice.n_atoms)

        app_vertex = self.app_vertex
        # Write the number of blocks of delays:
        spec.write_value(data=app_vertex.n_delay_stages)

        # write the delay per delay stage
        spec.write_value(data=app_vertex.delay_per_stage)

        # write whether to throw away spikes
        spec.write_value(data=int(app_vertex.drop_late_spikes))

        # Write the number of colour bits
        spec.write_value(data=app_vertex.n_colour_bits)



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence, Mapping
from pyNN.random import NumpyRNG
from spinn_front_end_common.interface.ds import DataType
from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.models.neuron.implementations import (
    AbstractNeuronImpl, ModelParameter)
from spynnaker.pyNN.data.spynnaker_data_view import SpynnakerDataView
from spynnaker.pyNN.random_distribution import RandomDistribution

V_INIT = "v_init"
V_RESET = "v_reset"
TAU = "tau"
TAU_REFRAC = "tau_refrac"
TIMESTEP = "timestep"
BIAS = "bias"
REFRACT_INIT = "refract_init"
SEED0 = "seed0"
SEED1 = "seed1"
SEED2 = "seed2"
SEED3 = "seed3"

UNITS = {
    TAU: "ms",
    BIAS: "mV",
    REFRACT_INIT: "ms",
}

MAX_INT = float(0xFFFFFFFF)


class NeuronImplStocExpStable(AbstractNeuronImpl):
    """ Stochastic neuron model with exponential threshold and instantaneous
        synapses, and voltage stays unless changed by input.
    """

    def __init__(self, v_init: ModelParameter, v_reset: ModelParameter,
                 tau: ModelParameter, tau_refrac: ModelParameter,
                 bias: ModelParameter, refract_init: ModelParameter,
                 seed: Optional[int]):
        """
        :param v_init:  :math:`V_{init}`
        :param v_reset: :math:`V_{thresh}`
        :param tau:
        :param tau_refrac:
        :param bias:
        :param refract_init:
        :param seed: Seed for the random distribution
        """
        self._v_init = v_init
        self._v_reset = v_reset
        self._tau = tau
        self._tau_refrac = tau_refrac
        self._bias = bias
        self._refract_init = refract_init
        self._random = RandomDistribution(
            "uniform", low=0, high=0xFFFFFFFF, rng=NumpyRNG(seed))

        self._struct = Struct([
            (DataType.S1615, V_INIT),
            (DataType.S1615, V_RESET),
            (DataType.U1616, TAU),
            (DataType.U1616, TAU_REFRAC),
            (DataType.U1616, TIMESTEP),
            (DataType.S1615, BIAS),
            (DataType.UINT32, REFRACT_INIT),
            (DataType.UINT32, SEED0),
            (DataType.UINT32, SEED1),
            (DataType.UINT32, SEED2),
            (DataType.UINT32, SEED3)])

    @property
    @overrides(AbstractNeuronImpl.structs)
    def structs(self) -> Sequence[Struct]:
        return [self._struct]

    @property
    @overrides(AbstractNeuronImpl.model_name)
    def model_name(self) -> str:
        return "StocExpStable"

    @property
    @overrides(AbstractNeuronImpl.binary_name)
    def binary_name(self) -> str:
        return "stoc_exp_stable.aplx"

    @overrides(AbstractNeuronImpl.get_global_weight_scale)
    def get_global_weight_scale(self) -> float:
        return 1.0

    @overrides(AbstractNeuronImpl.get_n_synapse_types)
    def get_n_synapse_types(self) -> int:
        return 2

    @overrides(AbstractNeuronImpl.get_synapse_id_by_target)
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        if target == "excitatory":
            return 0
        elif target == "inhibitory":
            return 1
        raise ValueError(f"Unknown target {target}")

    @overrides(AbstractNeuronImpl.get_synapse_targets)
    def get_synapse_targets(self) -> Sequence[str]:
        return ["excitatory", "inhibitory"]

    @overrides(AbstractNeuronImpl.get_recordable_variables)
    def get_recordable_variables(self) -> Sequence[str]:
        return ["v", "ex_input", "in_input", "prob"]

    @overrides(AbstractNeuronImpl.get_recordable_data_types)
    def get_recordable_data_types(self) -> Mapping[str, DataType]:
        return {"v": DataType.S1615,
                "ex_input": DataType.S1615, "in_input": DataType.S1615,
                "prob": DataType.U032}

    @overrides(AbstractNeuronImpl.get_recordable_units)
    def get_recordable_units(self, variable: str) -> str:
        # TODO: Update with the appropriate units for variables
        if variable in ("v", "ex_input", "in_input"):
            return "mV"
        elif variable == "prob":
            return ""
        raise ValueError(f"Unknown variable {variable}")

    @overrides(AbstractNeuronImpl.get_recordable_variable_index)
    def get_recordable_variable_index(self, variable: str) -> int:
        if variable == "v":
            return 0
        if variable == "ex_input":
            return 1
        if variable == "in_input":
            return 2
        if variable == "prob":
            return 3
        raise ValueError(f"Unknown variable {variable}")

    @overrides(AbstractNeuronImpl.is_recordable)
    def is_recordable(self, variable: str) -> bool:
        # TODO: Update to identify variables that can be recorded
        return variable in ("v", "ex_input", "in_input", "prob")

    @overrides(AbstractNeuronImpl.add_parameters)
    def add_parameters(self, parameters: RangeDictionary) -> None:
        parameters[V_RESET] = self._v_reset
        parameters[TAU] = self._tau
        parameters[TAU_REFRAC] = self._tau_refrac
        parameters[TIMESTEP] = SpynnakerDataView.get_simulation_time_step_ms()
        parameters[BIAS] = self._bias

    @overrides(AbstractNeuronImpl.add_state_variables)
    def add_state_variables(self, state_variables: RangeDictionary) -> None:
        state_variables[V_INIT] = self._v_init
        state_variables[REFRACT_INIT] = self._refract_init
        state_variables[SEED0] = self._random
        state_variables[SEED1] = self._random
        state_variables[SEED2] = self._random
        state_variables[SEED3] = self._random

    @overrides(AbstractNeuronImpl.get_units)
    def get_units(self, variable: str) -> str:
        return UNITS[variable]

    @property
    @overrides(AbstractNeuronImpl.is_conductance_based)
    def is_conductance_based(self) -> bool:
        return False



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Any, Callable, cast, Dict, Iterable, Iterator, List, Optional, overload,
    Tuple, Union)
from pyNN.random import RandomDistribution
from spinn_utilities.helpful_functions import is_singleton
from spinn_utilities.ranged.abstract_sized import Selector


class ParameterHolder(object):
    """
    Holds a set of parameters and state variables to be returned in a
    PyNN-specific format.
    """

    __slots__ = (
        # A list of items of data that are to be present in each element
        "__data_items_to_return",

        "__single_key",

        # Function call to get the values
        "__get_call",

        # The merged parameters formed just before the data is read
        "__data_items",

        # A selector to use if requested
        "__selector")

    def __init__(
            self, data_items_to_return: Union[str, Iterable[str]],
            get_call: Callable[[str, Selector], Union[
                List[float], RandomDistribution]],
            selector: Selector = None):
        """
        :param data_items_to_return: A list of data fields to be returned
        :param get_call: A function to call to read a value
        :param selector: a description of the subrange to accept,
            or `None` for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids`
        """
        self.__data_items_to_return: Union[str, Tuple[str, ...]]
        if isinstance(data_items_to_return, str):
            self.__data_items_to_return = data_items_to_return
            self.__single_key: Optional[str] = data_items_to_return
        else:
            self.__data_items_to_return = tuple(data_items_to_return)
            self.__single_key = None
        self.__get_call = get_call
        self.__data_items: Optional[Dict[str, List[float]]] = None
        self.__selector = selector

    def _safe_read_values(self, parameter: str) -> Union[List[float], float]:
        values = self.__get_call(parameter, self.__selector)

        # The values must be a single item, a list or a random distribution;
        # if a random distribution we must not have generated yet!
        if isinstance(values, RandomDistribution):
            raise ValueError(
                "Although it is possible to request the values"
                " before the simulation has run, it is not possible to read"
                " those values until after the simulation has run.  Please run"
                f" the simulation before reading {parameter}.")

        if self.__selector is not None and is_singleton(self.__selector):
            return values[0]
        return values

    def _get_data_items(self) -> Dict[str, List[float]]:
        """
        Merges the parameters and values in to the final data items
        """
        # If there are already merged connections cached, return those
        if self.__data_items is not None:
            return self.__data_items

        # If there is just one item to return, return the values stored
        if is_singleton(self.__data_items_to_return):
            key = cast(str, self.__data_items_to_return)
            self.__data_items = {
                key: cast(List[float], self._safe_read_values(key))}
            return self.__data_items

        # If there are multiple items to return, form a list
        self.__data_items = {
            param: cast(List[float], self._safe_read_values(param))
            for param in self.__data_items_to_return}

        return self.__data_items

    @overload
    def __getitem__(self, s: int) -> float:
        ...

    @overload
    def __getitem__(self, s: str) -> List[float]:
        ...

    def __getitem__(self, s: Union[int, str]) -> Union[float, List[float]]:
        data = self._get_data_items()
        if self.__single_key is not None:
            if not isinstance(s, int):
                raise KeyError("As there is only one array held "
                               "only int parameter are valid")
            return data[self.__single_key][s]
        if not isinstance(s, str):
            raise KeyError("As multiple arrays held "
                           "only str parameter are valid")
        return data[s]

    def __len__(self) -> int:
        data = self._get_data_items()
        if self.__single_key is not None:
            return len(data[self.__single_key])
        return len(data)

    def __iter__(self) -> Iterator:
        data = self._get_data_items()
        if self.__single_key is not None:
            return iter(data[self.__single_key])
        return iter(data)

    def __str__(self) -> str:
        data = self._get_data_items()
        if self.__single_key is not None:
            return str(data[self.__single_key])
        return str(data)

    def __repr__(self) -> str:
        data = self._get_data_items()
        if self.__single_key is not None:
            return repr(data[self.__single_key])
        return repr(data)

    def __contains__(self, item: Union[str, int]) -> bool:
        data = self._get_data_items()
        if self.__single_key is not None:
            return item in data[self.__single_key]
        return item in data

    def __eq__(self, other: Any) -> bool:
        data = self._get_data_items()
        if self.__single_key is not None:
            return data[self.__single_key] == other
        return data == other

    def __hash__(self) -> int:
        data = self._get_data_items()
        if self.__single_key is not None:
            return hash(data[self.__single_key])
        return hash(data)

    def keys(self) -> Iterable[str]:
        """
        :returns: The names of the data
        """
        data = self._get_data_items()
        return data.keys()

    def values(self) -> Iterable[List[float]]:
        """
        :returns: The values of the data
        """
        data = self._get_data_items()
        return data.values()

    def items(self) -> Iterable[Tuple[str, List[float]]]:
        """
        :returns: Iterable of the names and matching values of the data
        """
        data = self._get_data_items()
        return data.items()



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Mapping, Tuple
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.graphs.application import ApplicationGraph
from spynnaker.pyNN.models.neuron import ConnectionHolder
from spynnaker.pyNN.models.neural_projections import (
    ProjectionApplicationEdge, SynapseInformation)


class SpYNNakerConnectionHolderGenerator(object):
    """
    Sets up connection holders for reports to use.
    """

    def __call__(self, application_graph: ApplicationGraph) -> Mapping[
            Tuple[ProjectionApplicationEdge, SynapseInformation],
            ConnectionHolder]:
        """
        :param application_graph: application graph
        :return:
            the set of connection holders for after data specification
            generation
        """
        progress = ProgressBar(
            application_graph.n_outgoing_edge_partitions,
            "Generating connection holders for reporting connection data.")

        data_holders: Dict[
            Tuple[ProjectionApplicationEdge, SynapseInformation],
            ConnectionHolder] = dict()
        for partition in progress.over(
                application_graph.outgoing_edge_partitions):
            for edge in partition.edges:
                # add pre run generators so that reports can extract without
                # going to machine.
                if isinstance(edge, ProjectionApplicationEdge):
                    # build connection holders
                    self._generate_holder_for_edge(edge, data_holders)

        # return the two holders
        return data_holders

    @staticmethod
    def _generate_holder_for_edge(
            edge: ProjectionApplicationEdge, data_holders: Dict[
                Tuple[ProjectionApplicationEdge, SynapseInformation],
                ConnectionHolder]) -> None:
        # build connection holders
        connection_holder = ConnectionHolder(
            None, True, edge.pre_vertex.n_atoms, edge.post_vertex.n_atoms)

        for synapse_information in edge.synapse_information:
            synapse_information.add_pre_run_connection_holder(
                connection_holder)
            # store for the report generations
            data_holders[edge, synapse_information] = connection_holder



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import numpy
from numpy import uint16, uint32
from numpy.typing import NDArray

# Default value of fixed-point one for STDP
STDP_FIXED_POINT_ONE = (1 << 11)


def float_to_fixed(value: float) -> int:
    """
    :param value:
    :returns: value in fixed point format
    """
    return int(round(float(value) * STDP_FIXED_POINT_ONE))


def get_exp_lut_array(time_step: float, time_constant: float,
                      shift: int = 0) -> NDArray[uint32]:
    """
    :param time_step:
    :param time_constant:
    :param shift:
    :returns: Data as a lookup table
    """
    # Compute the actual exponential decay parameter
    # NB: lambda is a reserved word in Python
    l_ambda = time_step / float(time_constant)

    # Compute the size of the array, which must be a multiple of 2
    size = math.log(STDP_FIXED_POINT_ONE) / l_ambda
    size, extra = divmod(size / (1 << shift), 2)
    size = (int(size) + (extra > 0)) * 2

    # Fill out the values in the array
    a = numpy.exp((numpy.arange(size) << shift) * -l_ambda)
    a = numpy.floor(a * STDP_FIXED_POINT_ONE)

    # Concatenate with the header
    header = numpy.array([len(a), shift], dtype=uint16)
    return numpy.concatenate((header, a.astype(uint16))).view(uint32)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import select
import socket
from typing import Callable, Optional, TypeVar
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.ping import Ping
from spinnman.connections.abstract_classes import Listenable, Connection
from spinnman.utilities.socket_utils import (
    get_tcp_socket, connect_socket, get_socket_address, resolve_host,
    receive_message, send_message)
from spinn_front_end_common.utilities.constants import BYTES_PER_KB
#: :meta private:
T = TypeVar("T")

logger = FormatAdapter(logging.getLogger(__name__))
# A set of connections that have already been made
_existing_connections = dict()


def get_pushbot_wifi_connection(
        remote_host: str, remote_port: int = 56000) -> "PushBotWIFIConnection":
    """
    Get an existing connection to a PushBot, or make a new one.

    :param remote_host: The IP address of the PushBot
    :param remote_port: The port number of the PushBot (default 56000)
    :returns: Connection to the Pushbot
    """
    key = (remote_host, remote_port)
    if key not in _existing_connections:
        _existing_connections[key] = \
            PushBotWIFIConnection(remote_host, remote_port)
    return _existing_connections[key]


class PushBotWIFIConnection(Connection, Listenable):
    """
    A connection to a PushBot via Wi-Fi.
    """
    __slots__ = (
        "__local_ip_address",
        "__local_port",
        "__remote_ip_address",
        "__remote_port",
        "__socket")

    RECV_SIZE = 1 * BYTES_PER_KB

    def __init__(self, remote_host: str, remote_port: int = 56000):
        """
        :param remote_host: The IP address of the PushBot
        :param remote_port: The port number of the PushBot (default 56000)
        :raise ~spinnman.exceptions.SpinnmanIOException:
            If there is an error setting up the communication channel
        """
        # Create a TCP Socket
        self.__socket = get_tcp_socket()

        # Get the port to connect to
        self.__remote_port = int(remote_port)

        # Get the host to connect to
        self.__remote_ip_address = resolve_host(remote_host)

        logger.info("Trying to connect to the PushBot via Wi-Fi")
        # Connect the socket
        connect_socket(
            self.__socket, self.__remote_ip_address, self.__remote_port)
        logger.info("Succeeded in connecting to PushBot via Wi-Fi")

        # Get the details of where the socket is connected
        self.__local_ip_address, self.__local_port = get_socket_address(
            self.__socket)

    def is_connected(self) -> bool:
        """
        See
        :py:meth:`~spinnman.connections.Connection.is_connected

        :returns: True if ping works
        """
        # check if machine is active and on the network
        for _ in range(5):  # Try up to five times...
            # ping the remote address
            if Ping.ping(self.__remote_ip_address) == 0:
                # ping worked
                return True

        # If the ping fails this number of times, the host cannot be contacted
        return False

    @property
    def local_ip_address(self) -> str:
        """
        The local IP address to which the connection is bound,
        as a dotted string, e.g. `0.0.0.0`.
        """
        return self.__local_ip_address

    @property
    def local_port(self) -> int:
        """
        The local port to which the connection is bound.
        """
        return self.__local_port

    @property
    def remote_ip_address(self) -> Optional[str]:
        """
        The remote IP address to which the connection is connected,
        as a dotted string, or `None` if not connected remotely.
        """
        return self.__remote_ip_address

    @property
    def remote_port(self) -> int:
        """
        The remote port to which the connection is connected,
        or `None` if not connected remotely.

        """
        return self.__remote_port

    def receive(self, timeout: Optional[float] = None) -> bytes:
        """
        Receive data from the connection

        :param timeout: The timeout, or `None` to wait forever
        :return: The data received
        :raise SpinnmanTimeoutException:
            If a timeout occurs before any data is received
        :raise ~spinnman.exceptions.SpinnmanIOException:
            If an error occurs receiving the data
        """
        return receive_message(self.__socket, timeout, self.RECV_SIZE)

    def send(self, data: bytes) -> None:
        """
        Send data down this connection

        :param data: The data to be sent
        :raise ~spinnman.exceptions.SpinnmanIOException:
            If there is an error sending the data
        """
        send_message(self.__socket, data)

    def close(self) -> None:
        """
        See
        :py:meth:`spinnman.connections.Connection.close`.
        """
        try:
            self.__socket.shutdown(socket.SHUT_WR)
        except Exception:  # pylint: disable=broad-except
            pass
        self.__socket.close()

    @overrides(Listenable.is_ready_to_receive)
    def is_ready_to_receive(self, timeout: float = 0) -> bool:
        return bool(select.select([self.__socket], [], [], timeout)[0])

    @overrides(Listenable.get_receive_method)
    def get_receive_method(self) -> Callable[[], bytes]:
        return self.receive



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Sequence, Tuple, TYPE_CHECKING
from numpy.typing import NDArray
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from pacman.model.graphs.common import Slice
if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation


class AbstractGenerateConnectorOnHost(object, metaclass=AbstractBase):
    """
    A connector that can be generated on host.
    """

    # Mix-in class, so don't add anything here!
    __slots__: Tuple[str, ...] = ()

    @abstractmethod
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        """
        Create a synaptic block from the data.

        :param post_slices:
        :param post_vertex_slice:
        :param synapse_type:
        :param synapse_info:
        :returns:
            The synaptic matrix data to go to the machine, as a Numpy array
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any, Dict, TYPE_CHECKING
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
if TYPE_CHECKING:
    from spynnaker.pyNN.connections import SpynnakerLiveSpikesConnection
    from .abstract_ethernet_translator import AbstractEthernetTranslator


class AbstractEthernetSensor(object, metaclass=AbstractBase):
    """
    An Ethernet-connected device that can send events (spikes) to SpiNNaker
    via a Spike Injector.
    """
    __slots__ = ()

    @abstractmethod
    def get_n_neurons(self) -> int:
        """
        :returns: The number of neurons that will be sent out by the device.
        """
        raise NotImplementedError

    @abstractmethod
    def get_injector_parameters(self) -> Dict[str, Any]:
        """
        :returns:
           The parameters of the Spike Injector to use with this device.
        """
        raise NotImplementedError

    @abstractmethod
    def get_injector_label(self) -> str:
        """
        :returns: the label to give to the Spike Injector.
        """
        raise NotImplementedError

    @abstractmethod
    def get_translator(self) -> AbstractEthernetTranslator:
        """
        :returns: A translator of multicast commands to Ethernet commands.
        """
        raise NotImplementedError

    @abstractmethod
    def get_database_connection(self) -> SpynnakerLiveSpikesConnection:
        """
        :returns: A Database Connection instance that this device uses to
             inject packets.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import (Dict, List, Final, Optional, Sequence, Tuple, Union,
                    TYPE_CHECKING)

import numpy
from numpy import floating, integer, ndarray, uint32
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from pyNN.random import RandomDistribution
from pyNN.space import Space

from spinn_utilities.overrides import overrides

from pacman.model.graphs import AbstractVertex
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.types import (
    DELAYS, WEIGHTS_DELAYS, WEIGHTS)

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)


_TWOD: Final['TypeAlias'] = Union[List[int], Tuple[int, int]]
_KERNAL: Final['TypeAlias'] = Union[
    float, int, List[float], NDArray[numpy.floating], RandomDistribution]

HEIGHT, WIDTH = 0, 1
N_KERNEL_PARAMS = 9


class ConvolutionKernel(ndarray):
    """
    Thin wrapper around a numpy array.
    """


def shape2word(
        short1: Union[int, integer], short2: Union[int, integer]) -> uint32:
    """
    Combines two short values into 1 int by shifting the first 16 places

    :param short1: first 2 byte value
    :param short2: second 2 bytes value
    :returns: Combination of the two values with the first shifted 16 places
    """
    return uint32(((uint32(short2) & 0xFFFF) << 16)
                  | (uint32(short1) & 0xFFFF))


class KernelConnector(AbstractGenerateConnectorOnMachine,
                      AbstractGenerateConnectorOnHost):
    """
    Where the pre- and post-synaptic populations are considered as a 2D
    array. Connect every post(row, column) neuron to many
    pre(row, column, kernel)
    through a (kernel) set of weights and/or delays.
    """
    __slots__ = (
        "_kernel_w", "_kernel_h",
        "_hlf_k_w", "_hlf_k_h",
        "_pre_w", "_pre_h",
        "_post_w", "_post_h",
        "_pre_start_w", "_pre_start_h",
        "_post_start_w", "_post_start_h",
        "_pre_step_w", "_pre_step_h",
        "_post_step_w", "_post_step_h",
        "_krn_weights", "_krn_delays", "_shape_common",
        "_common_w", "_common_h",
        "_shape_pre", "_shape_post",
        "_post_as_pre")

    def __init__(
            self, shape_pre: _TWOD, shape_post: _TWOD, shape_kernel: _TWOD,
            weight_kernel: Optional[_KERNAL] = None,
            delay_kernel: Optional[_KERNAL] = None,
            shape_common: Optional[_TWOD] = None,
            pre_sample_steps_in_post: Optional[_TWOD] = None,
            pre_start_coords_in_post: Optional[_TWOD] = None,
            post_sample_steps_in_pre: Optional[_TWOD] = None,
            post_start_coords_in_pre: Optional[_TWOD] = None,
            safe: bool = True, space: Optional[Space] = None,
            verbose: bool = False, callback: None = None):
        """
        :param shape_pre:
            2D shape of the pre-population (rows/height, columns/width, usually
            the input image shape)
        :param shape_post:
            2D shape of the post-population (rows/height, columns/width)
        :param shape_kernel:
            2D shape of the kernel (rows/height, columns/width)
        :param weight_kernel: (optional)
            2D matrix of size shape_kernel describing the weights
        :param delay_kernel: (optional)
            2D matrix of size shape_kernel describing the delays
        :param shape_common: (optional)
            2D shape of common coordinate system (for both pre- and post-,
            usually the input image sizes)
        :param pre_sample_steps_in_post: (optional)
            Sampling steps/jumps for pre-population <=> (stepX, stepY)
        :param pre_start_coords_in_post: (optional)
            Starting row/column for pre-population sampling <=> (offX, offY)
        :param post_sample_steps_in_pre: (optional)
            Sampling steps/jumps for post-population <=> (stepX, stepY)
        :param post_start_coords_in_pre: (optional)
            Starting row/column for post-population sampling <=> (offX, offY)
        :param safe:
            Whether to check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param space:
            Currently ignored; for future compatibility.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe=safe, callback=callback, verbose=verbose)
        assert space is None, "non-None space unsupported"

        # Get the kernel size
        self._kernel_w = shape_kernel[WIDTH]
        self._kernel_h = shape_kernel[HEIGHT]

        # The half-value used here indicates the half-way array position
        self._hlf_k_w = shape_kernel[WIDTH] // 2
        self._hlf_k_h = shape_kernel[HEIGHT] // 2

        # Cache values for the pre and post sizes
        self._pre_w = shape_pre[WIDTH]
        self._pre_h = shape_pre[HEIGHT]
        self._post_w = shape_post[WIDTH]
        self._post_h = shape_post[HEIGHT]

        # Get the starting coordinates and step sizes
        # (or defaults if not given)
        if pre_start_coords_in_post is None:
            self._pre_start_w = 0
            self._pre_start_h = 0
        else:
            self._pre_start_w = pre_start_coords_in_post[WIDTH]
            self._pre_start_h = pre_start_coords_in_post[HEIGHT]

        if post_start_coords_in_pre is None:
            self._post_start_w = 0
            self._post_start_h = 0
        else:
            self._post_start_w = post_start_coords_in_pre[WIDTH]
            self._post_start_h = post_start_coords_in_pre[HEIGHT]

        if pre_sample_steps_in_post is None:
            self._pre_step_w = 1
            self._pre_step_h = 1
        else:
            self._pre_step_w = pre_sample_steps_in_post[WIDTH]
            self._pre_step_h = pre_sample_steps_in_post[HEIGHT]

        if post_sample_steps_in_pre is None:
            self._post_step_w = 1
            self._post_step_h = 1
        else:
            self._post_step_w = post_sample_steps_in_pre[WIDTH]
            self._post_step_h = post_sample_steps_in_pre[HEIGHT]

        # Make sure the supplied values are in the correct format
        self._krn_weights = self.__get_kernel_vals(weight_kernel)
        self._krn_delays = self.__get_kernel_vals(delay_kernel)

        self._shape_common = \
            shape_pre if shape_common is None else shape_common
        self._common_w = self._shape_common[WIDTH]
        self._common_h = self._shape_common[HEIGHT]
        self._shape_pre = shape_pre
        self._shape_post = shape_post

        # Create storage for later
        self._post_as_pre: Dict[
            Slice, Tuple[NDArray[integer], NDArray[integer]]] = {}

    def __to_post_coords(
            self, post_vertex_slice: Slice) -> Tuple[
                NDArray[integer], NDArray[integer]]:
        """
        Get a list of possible post-slice coordinates.

        :param post_vertex_slice:
        """
        post = numpy.arange(
            post_vertex_slice.lo_atom, post_vertex_slice.hi_atom + 1,
            dtype=uint32)
        return numpy.divmod(post, self._post_w)

    def __map_to_pre_coords(
            self, post_r: NDArray[integer], post_c: NDArray[integer]) -> Tuple[
                NDArray[integer], NDArray[integer]]:
        """
        Get a map from post to pre-population coordinates.
        """
        return (self._post_start_h + post_r * self._post_step_h,
                self._post_start_w + post_c * self._post_step_w)

    def __post_as_pre(self, post_vertex_slice: Slice) -> Tuple[
            NDArray[integer], NDArray[integer]]:
        """
        Write post-population coordinates as pre-population coordinates.
        """
        # directly as the cache index
        if post_vertex_slice not in self._post_as_pre:
            post_r, post_c = self.__to_post_coords(post_vertex_slice)
            self._post_as_pre[post_vertex_slice] = \
                self.__map_to_pre_coords(post_r, post_c)
        return self._post_as_pre[post_vertex_slice]

    def __pre_as_post(self, pre_r: int, pre_c: int) -> Tuple[int, int]:
        """
        Write pre-population coordinates as post-population coordinates.
        """
        r = ((pre_r - self._pre_start_h - 1) // self._pre_step_h) + 1
        c = ((pre_c - self._pre_start_w - 1) // self._pre_step_w) + 1
        return (r, c)

    def __get_kernel_vals(self, values: Optional[Union[
            _KERNAL, WEIGHTS_DELAYS]]) -> Optional[ConvolutionKernel]:
        """
        Convert kernel values given into the correct format.
        """
        if values is None:
            return None
        if isinstance(values, list):
            values = numpy.asarray(values)
        krn_size = self._kernel_h * self._kernel_w
        krn_shape = (self._kernel_h, self._kernel_w)
        if isinstance(values, RandomDistribution):
            return numpy.array(values.next(krn_size)).reshape(krn_shape).view(
                ConvolutionKernel)
        elif numpy.isscalar(values):
            return numpy.full(krn_shape, values).view(ConvolutionKernel)
        elif ((isinstance(values, numpy.ndarray) or
               isinstance(values, ConvolutionKernel)) and
              values.shape[HEIGHT] == self._kernel_h and
              values.shape[WIDTH] == self._kernel_w):
            return values.view(ConvolutionKernel)
        raise SpynnakerException(
            "Error generating KernelConnector values; if you have supplied "
            "weight and/or delay kernel then ensure they are the same size "
            "as specified by the shape kernel values (height: "
            f"{self._kernel_h} and width: {self._kernel_w}).")

    def __compute_statistics(
            self, weights: Optional[WEIGHTS],
            delays: Optional[DELAYS], post_vertex_slice: Slice,
            n_pre_neurons: int) -> Tuple[
                int, NDArray[uint32], NDArray[uint32], NDArray[floating],
                NDArray[floating]]:
        """
        Compute the relevant information required for the connections.
        """
        # If __compute_statistics is called more than once, there's
        # no need to get the user-supplied weights and delays again
        if self._krn_weights is None:
            self._krn_weights = self.__get_kernel_vals(weights)
        if self._krn_delays is None:
            self._krn_delays = self.__get_kernel_vals(delays)
        assert self._krn_weights is not None
        assert self._krn_delays is not None

        post_as_pre_r, post_as_pre_c = self.__post_as_pre(post_vertex_slice)
        coords: Dict[int, List[int]] = {}
        hh, hw = self._hlf_k_h, self._hlf_k_w
        all_pre_ids: List[int] = []
        all_post_ids: List[int] = []
        all_delays: List[NDArray[floating]] = []
        all_weights: List[NDArray[floating]] = []
        count = 0
        post_lo = post_vertex_slice.lo_atom

        # Loop over pre-vertices
        for pre_idx in range(n_pre_neurons):
            pre_r, pre_c = divmod(pre_idx, self._pre_w)
            coords[pre_idx] = []

            # Test whether the coordinates should be included based on the
            # step function (in the pre) and skip if not
            if not (((pre_r - self._pre_start_h) % self._pre_step_h == 0) and
                    ((pre_c - self._pre_start_w) % self._pre_step_w == 0)):
                continue

            # Loop over post-vertices
            for post_idx in range(
                    post_vertex_slice.lo_atom, post_vertex_slice.hi_atom + 1):
                # convert to common coordinate system
                pac_r = post_as_pre_r[post_idx - post_lo]
                pac_c = post_as_pre_c[post_idx - post_lo]

                # now convert common to pre coordinates
                pap_r, pap_c = self.__pre_as_post(pac_r, pac_c)

                # Obtain coordinates to test against kernel sizes
                dr = pap_r - pre_r
                kr = hh - dr
                dc = pap_c - pre_c
                kc = hw - dc

                if 0 <= kr < self._kernel_h and 0 <= kc < self._kernel_w:
                    if post_idx in coords[pre_idx]:
                        continue
                    coords[pre_idx].append(post_idx)

                    # Store weights, delays and pre/post ids
                    w = self._krn_weights[kr, kc]
                    d = self._krn_delays[kr, kc]

                    count += 1

                    all_pre_ids.append(pre_idx)
                    all_post_ids.append(post_idx)
                    all_delays.append(d)
                    all_weights.append(w)

        # Now the loop is complete, return relevant data
        return (count, numpy.array(all_post_ids, dtype=uint32),
                numpy.array(all_pre_ids, dtype=uint32),
                numpy.array(all_delays), numpy.array(all_weights))

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        # Use the kernel delays if user has supplied them
        if self._krn_delays is not None:
            return numpy.max(self._krn_delays)

        # I think this is overestimated, but not by much
        n_conns = (
            self._pre_w * self._pre_h * self._kernel_w * self._kernel_h)

        # if not then use the values that came in
        return self._get_delay_maximum(
            synapse_info.delays, n_conns, synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        # Use the kernel delays if user has supplied them
        if self._krn_delays is not None:
            return numpy.min(self._krn_delays)

        # I think this is overestimated, but not by much
        n_conns = (
            self._pre_w * self._pre_h * self._kernel_w * self._kernel_h)

        # if not then use the values that came in
        return self._get_delay_minimum(
            synapse_info.delays, n_conns, synapse_info)

    @overrides(AbstractConnector.get_delay_variance)
    def get_delay_variance(self, delays: DELAYS,
                           synapse_info: SynapseInformation) -> float:
        if self._krn_delays is not None:
            return float(numpy.var(self._krn_delays))
        return super().get_delay_variance(delays, synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        return numpy.clip(self._kernel_h * self._kernel_w, 0, n_post_atoms)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        return numpy.clip(self._kernel_h * self._kernel_w, 0, 255)

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        # Use the kernel weights if user has supplied them
        if self._krn_weights is not None:
            return numpy.max(self._krn_weights)

        # I think this is overestimated, but not by much
        n_conns = self._pre_w * self._pre_h * self._kernel_w * self._kernel_h
        return self._get_weight_maximum(
            synapse_info.weights, n_conns, synapse_info)

    @overrides(AbstractConnector.get_weight_mean)
    def get_weight_mean(self, weights: WEIGHTS,
                        synapse_info: SynapseInformation) -> float:
        # Use the kernel weights if user has supplied them
        if self._krn_weights is not None:
            return float(numpy.mean(self._krn_weights))
        return super().get_weight_mean(weights, synapse_info)

    @overrides(AbstractConnector.get_weight_variance)
    def get_weight_variance(self, weights: WEIGHTS,
                            synapse_info: SynapseInformation) -> float:
        # Use the kernel weights if user has supplied them
        if self._krn_weights is not None:
            return float(numpy.var(self._krn_weights))
        return super().get_weight_variance(weights, synapse_info)

    def __repr__(self) -> str:
        return \
            f"KernelConnector(shape_kernel[{self._kernel_w},{self._kernel_h}])"

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        (n_connections, all_post, all_pre_in_range, all_pre_in_range_delays,
         all_pre_in_range_weights) = self.__compute_statistics(
            synapse_info.weights, synapse_info.delays, post_vertex_slice,
            synapse_info.n_pre_neurons)

        syn_dtypes = AbstractConnector.NUMPY_SYNAPSES_DTYPE

        if n_connections <= 0:
            return numpy.zeros(0, dtype=syn_dtypes)

        # 0 for excitatory, 1 for inhibitory
        syn_type = numpy.array(all_pre_in_range_weights < 0)
        block = numpy.zeros(n_connections, dtype=syn_dtypes)
        block["source"] = all_pre_in_range
        block["target"] = all_post
        block["weight"] = all_pre_in_range_weights
        block["delay"] = all_pre_in_range_delays
        block["synapse_type"] = syn_type.astype('uint8')
        return block

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.KERNEL_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        data = numpy.array([
            shape2word(self._common_w, self._common_h),
            shape2word(self._pre_w, self._pre_h),
            shape2word(self._post_w, self._post_h),
            shape2word(self._pre_start_w, self._pre_start_h),
            shape2word(self._post_start_w, self._post_start_h),
            shape2word(self._pre_step_w, self._pre_step_h),
            shape2word(self._post_step_w, self._post_step_h),
            shape2word(self._kernel_w, self._kernel_h),
            shape2word(int(self._krn_weights is not None),
                       int(self._krn_delays is not None))], dtype=uint32)
        extra_data = []
        if self._krn_weights is not None:
            extra_data.append(DataType.S1615.encode_as_numpy_int_array(
                self._krn_weights.flatten()))
        if self._krn_delays is not None:
            extra_data.append(DataType.S1615.encode_as_numpy_int_array(
                self._krn_delays.flatten()))

        if extra_data:
            return numpy.concatenate((data, *extra_data))
        return data

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        size = N_KERNEL_PARAMS * BYTES_PER_WORD
        if self._krn_weights is not None:
            size += sum(len(x) for x in self._krn_weights) * BYTES_PER_WORD
        if self._krn_delays is not None:
            size += sum(len(x) for x in self._krn_delays) * BYTES_PER_WORD
        return size

    @overrides(AbstractGenerateConnectorOnMachine.get_connected_vertices)
    def get_connected_vertices(
            self, s_info: SynapseInformation, source_vertex: ApplicationVertex,
            target_vertex: ApplicationVertex) -> Sequence[
                Tuple[MachineVertex, Sequence[AbstractVertex]]]:
        src_splitter = source_vertex.splitter
        return [
            (t_vert,
             [s_vert for s_vert in src_splitter.get_out_going_vertices(
                 s_info.partition_id) if self.__connects(s_vert, t_vert)])
            for t_vert in target_vertex.splitter.get_in_coming_vertices(
                s_info.partition_id)]

    def __connects(self, src_machine_vertex: MachineVertex,
                   dest_machine_vertex: MachineVertex) -> bool:
        # If the pre- and post-slices are not 2-dimensional slices, we have
        # to let them pass
        pre_slice = src_machine_vertex.vertex_slice
        post_slice = dest_machine_vertex.vertex_slice
        if (pre_slice.shape is None or len(pre_slice.shape) != 2 or
                post_slice.shape is None or len(post_slice.shape) != 2):
            return True

        pre_slice_x = pre_slice.get_slice(0)
        pre_slice_y = pre_slice.get_slice(1)
        post_slice_x = post_slice.get_slice(0)
        post_slice_y = post_slice.get_slice(1)

        min_pre_x = post_slice_x.start - self._hlf_k_w
        max_pre_x = (post_slice_x.stop + self._hlf_k_w) - 1
        min_pre_y = post_slice_y.start - self._hlf_k_h
        max_pre_y = (post_slice_y.stop + self._hlf_k_h) - 1

        # No part of the pre square overlaps the post-square, don't connect
        if (pre_slice_x.stop <= min_pre_x or
                pre_slice_x.start > max_pre_x or
                pre_slice_y.stop <= min_pre_y or
                pre_slice_y.start > max_pre_y):
            return False

        # Otherwise, they do
        return True

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        pre = application_edge.pre_vertex
        post = application_edge.post_vertex
        if len(pre.atoms_shape) != 1 or len(post.atoms_shape) != 1:
            raise ConfigurationException(
                "The Kernel Connector is designed to work with 1D vertices")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Union


class RunningStats(object):
    """
    Keeps running statistics.
    From: https://www.johndcook.com/blog/skewness_kurtosis/
    """
    __slots__ = ("__mean", "__mean_2", "__n_items")

    def __init__(self) -> None:
        self.__mean = 0.0
        self.__mean_2 = 0.0
        self.__n_items = 0

    def add_item(self, x: Union[int, float]) -> None:
        """
        Adds an item to the running statistics.

        :param x: The item to add
        """
        old_n_items = self.__n_items
        self.__n_items += 1

        delta = x - self.__mean
        delta_n = delta / self.__n_items
        term_1 = delta * delta_n * old_n_items

        self.__mean += delta_n
        self.__mean_2 += term_1

    def add_items(self, mean: float, variance: float, n_items: int) -> None:
        """
        Add a bunch of items (via their statistics).

        :param mean: The mean of the items to add.
        :param variance: The variance of the items to add.
        :param n_items: The number of items represented.
        """
        if n_items > 0:
            new_n_items = self.__n_items + n_items
            mean_2 = variance * (n_items - 1.0)

            delta = mean - self.__mean
            delta_2 = delta * delta
            new_mean = (((self.__n_items * self.__mean) + (n_items * mean)) /
                        new_n_items)
            new_mean_2 = (self.__mean_2 + mean_2 +
                          (delta_2 * self.__n_items * n_items) / new_n_items)

            self.__n_items = new_n_items
            self.__mean = new_mean
            self.__mean_2 = new_mean_2

    @property
    def n_items(self) -> int:
        """
        The number of items seen.
        """
        return self.__n_items

    @property
    def mean(self) -> float:
        """
        The mean of the items seen.
        """
        return self.__mean

    @property
    def variance(self) -> float:
        """
        The variance of the items seen.
        """
        if self.__n_items <= 1:
            return 0.0
        return self.__mean_2 / (self.__n_items - 1.0)

    @property
    def standard_deviation(self) -> float:
        """
        The population standard deviation of the items seen.
        """
        return math.sqrt(self.variance)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Union

import numpy
from numpy.typing import NDArray

from pyNN.recording.files import BaseFile, StandardTextFile

from .from_list_connector import FromListConnector


class FromFileConnector(FromListConnector):
    """
    Make connections according to a list read from a file.
    """
    __slots__ = ("_file", )

    def __init__(
            self, file: Union[str, BaseFile],  # @ReservedAssignment
            distributed: bool = False, safe: bool = True,
            callback: None = None, verbose: bool = False):
        """
        :param file:
            Either an open file object or the filename of a file containing a
            list of connections, in the format required by
            :py:class:`FromListConnector`.
            Column headers, if included in the file, must be specified using
            a list or tuple, e.g.::

                # columns = ["i", "j", "weight", "delay", "U", "tau_rec"]

            .. note::
                The header requires `#` at the beginning of the line.
        :param distributed:
            Basic pyNN says:

                if this is ``True``, then each node will read connections from
                a file called ``filename.x``, where ``x`` is the MPI rank. This
                speeds up loading connections for distributed simulations.

            .. note::
                Always leave this as ``False`` with sPyNNaker, which is not
                MPI-based.
        :param safe:
            Whether to check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        """
        self._file = file
        if isinstance(file, str):
            real_file = self.get_reader(file)
            try:
                conn_list = self._read_conn_list(real_file, distributed)
            finally:
                real_file.close()
        else:
            conn_list = self._read_conn_list(file, distributed)

        column_names = self.get_reader(self._file).get_metadata().get(
            'columns')
        if column_names is not None:
            column_names = [column for column in column_names
                            if column not in ("i", "j")]

        super().__init__(
            conn_list, safe=safe, verbose=verbose,
            column_names=column_names, callback=callback)

    def _read_conn_list(
            self, the_file: BaseFile, distributed: bool) -> NDArray:
        if not distributed:
            return the_file.read()
        filename = f"{os.path.basename(the_file.file)}."

        # This assumes it finds the files in the right order!
        conns = list()
        for found_file in os.listdir(os.path.dirname(the_file.file)):
            if found_file.startswith(filename):
                file_reader = self.get_reader(found_file)
                try:
                    conns.append(file_reader.read())
                finally:
                    file_reader.close()
        return numpy.concatenate(conns)

    def __repr__(self) -> str:
        return f"FromFileConnector({self._file})"

    def get_reader(self, file: str) -> BaseFile:  # @ReservedAssignment
        """
        Get a file reader object using the PyNN methods.

        :return: A pynn StandardTextFile or similar
        """
        return StandardTextFile(file, mode="r")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from collections import defaultdict
import sys
from typing import (
    Any, cast, Dict, Optional, Sequence, Tuple, TYPE_CHECKING, Union)
import numpy
from pyNN import descriptions
from spinn_utilities.classproperty import classproperty
from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod)
from spynnaker.pyNN.models.defaults import AbstractProvidesDefaults
from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.data import SpynnakerDataView
if TYPE_CHECKING:
    from spynnaker.pyNN.models.common.population_application_vertex import (
        PopulationApplicationVertex)


class AbstractPyNNModel(AbstractProvidesDefaults, metaclass=AbstractBase):
    """
    A Model that can be passed in to a Population object in PyNN.
    """

    __slots__ = ()

    # The maximum number of atoms per core for PyNN models
    _max_atoms_per_core: Dict[type, Optional[Tuple[int, ...]]] = defaultdict(
        lambda: None)

    @classmethod
    def verify_may_set(cls, param: str) -> None:
        """ If a Population has been created, this method will raise an
            exception; used to avoid setting global limits after a Population
            has been created.

        :param param:
            The parameter name that can be used to set the value being changed
            in the Population constructor instead.
        """
        SpynnakerDataView.check_user_can_act()
        if SpynnakerDataView.get_n_populations() == 0:
            return
        raise SpynnakerException(
            "Global set is not supported after a Population has been "
            "created. Either move it above the creation of all Populations "
            f"or provide {param} during the creation of each Population it "
            "applies to.")

    @classmethod
    def set_model_max_atoms_per_dimension_per_core(
            cls, n_atoms: Union[None, int, Tuple[int, ...]] = None) -> None:
        """
        Set the default maximum number of atoms per dimension per core for
        this model.  This can be overridden by the individual Population.
        The new value can be `None`, meaning that the maximum is the same as
        the number of atoms, an int, meaning all Populations of this model
        must have one dimension, or a tuple of *n* integers, meaning all
        Populations of this model must have *n* dimensions.
        If not all Populations of this model have the same number of
        dimensions, it is recommended to set this to `None` here and then
        set the maximum on each Population.

        :param n_atoms: The new maximum, or `None` for the largest possible
        """
        cls.verify_may_set(param="neurons_per_core")
        abs_max = cls.absolute_max_atoms_per_core
        if n_atoms is None:
            AbstractPyNNModel._max_atoms_per_core[cls] = None
        elif numpy.isscalar(n_atoms):
            if n_atoms > abs_max:
                raise SpynnakerException(
                    "The absolute maximum neurons per core for this"
                    f" model is {abs_max}")
            max_atoms_int: int = int(cast(int, n_atoms))
            AbstractPyNNModel._max_atoms_per_core[cls] = (max_atoms_int, )
        else:
            if numpy.prod(n_atoms) > abs_max:
                raise SpynnakerException(
                    "The absolute maximum sum of neurons per core for this"
                    f" model is {abs_max}")
            max_atoms_tuple: Tuple[int, ...] = cast(
                Tuple[int, ...],  n_atoms)
            AbstractPyNNModel._max_atoms_per_core[cls] = max_atoms_tuple

    @classmethod
    def get_model_max_atoms_per_dimension_per_core(cls) -> Tuple[int, ...]:
        """
        :returns:
            The maximum number of atoms per dimension per core for this model.
        """
        # If there is a stored value, use it
        max_stored = AbstractPyNNModel._max_atoms_per_core.get(cls)
        if max_stored is not None:
            return max_stored

        # Otherwise return the absolute maximum assuming 1D
        return (cls.absolute_max_atoms_per_core, )

    @classmethod
    def reset_all(cls) -> None:
        """
        Reset the maximum values for all classes.
        """
        AbstractPyNNModel._max_atoms_per_core.clear()

    @classproperty
    def absolute_max_atoms_per_core(  # pylint: disable=no-self-argument
            cls) -> int:
        """
        The absolute maximum number of atoms per core.

        This is an integer regardless of the number of dimensions
        in any vertex.

        :returns: The absolute maximum number of atoms per core.
        """
        return sys.maxsize

    @classmethod
    def get_parameter_names(cls) -> Sequence[str]:
        """
        :returns: The names of the parameters of the model.
        """
        return cls.default_parameters.keys()  # pylint: disable=no-member

    @classmethod
    def has_parameter(cls, name: str) -> bool:
        """
        :param name: The name of the parameter to check for
        :returns: True if the model has a parameter with the given name
        """
        return name in cls.default_parameters

    #: The default values for the parameters at the population level.
    #: These are parameters that can be passed in to the Population
    #: constructor in addition to the standard PyNN options.
    default_population_parameters: Dict[str, Any] = {}

    @classmethod
    def _get_default_population_parameters(cls) -> Dict[str, Any]:
        """
        Get the default population parameters.
        Slightly contorted to allow for overriding class variables.
        """
        return dict(cls.default_population_parameters)

    @abstractmethod
    def create_vertex(
            self, n_neurons: int, label: str) -> PopulationApplicationVertex:
        """
        Create a vertex for a population of the model.

        :param n_neurons: The number of neurons in the population
        :param label: The label to give to the vertex
        :return: An application vertex for the population
        """
        raise NotImplementedError

    @property
    def name(self) -> str:
        """
        The name of this model.
        """
        return self.__class__.__name__

    def describe(self, template: Optional[str] = 'modeltype_default.txt',
                 engine: str = 'default') -> str:
        """
        Returns a human-readable description of the population.

        The output may be customised by specifying a different template
        together with an associated template engine (see
        :mod:`pyNN.descriptions`).

        If ``template`` is ``None``, then a dictionary containing the template
        context will be returned.

        :param template: Template filename
        :param engine: Template substitution engine
        :returns: A human-readable description
        """
        context = {
            "name": self.name
        }
        return descriptions.render(engine, template, context)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.defaults import (
    AbstractProvidesDefaults, default_initial_values)
from spynnaker.pyNN.models.neuron.implementations import ModelParameter


# pylint: disable=wrong-spelling-in-docstring
class EIFConductanceAlphaPopulation(AbstractProvidesDefaults):
    """
    Exponential integrate and fire neuron with spike triggered and
    sub-threshold adaptation currents (isfa, ista reps.)

    .. warning::
        Not currently supported by the tool chain.
    """

    # noinspection PyPep8Naming
    @default_initial_values({"v", "w", "gsyn_exc", "gsyn_inh"})
    def __init__(
            self, tau_m: ModelParameter = 9.3667, cm: ModelParameter = 0.281,
            v_rest: ModelParameter = -70.6, v_reset: ModelParameter = -70.6,
            v_thresh: ModelParameter = -50.4, tau_syn_E: ModelParameter = 5.0,
            tau_syn_I: ModelParameter = 0.5, tau_refrac: ModelParameter = 0.1,
            i_offset: ModelParameter = 0.0, a: ModelParameter = 4.0,
            b: ModelParameter = 0.0805, v_spike: ModelParameter = -40.0,
            tau_w: ModelParameter = 144.0, e_rev_E: ModelParameter = 0.0,
            e_rev_I: ModelParameter = -80.0, delta_T: ModelParameter = 2.0,
            v: ModelParameter = -70.6, w: ModelParameter = 0.0,
            gsyn_exc: ModelParameter = 0.0, gsyn_inh: ModelParameter = 0.0):
        # pylint: disable=unused-argument, invalid-name
        """
        :param tau_m:
        :param cm:
        :param v_rest:
        :param v_reset:
        :param v_thresh:
        :param tau_syn_E:
        :param tau_syn_I:
        :param tau_refrac:
        :param i_offset:
        :param a:
        :param b:
        :param v_spike:
        :param tau_w:
        :param e_rev_E:
        :param e_rev_I:
        :param delta_T:
        :param v:
        :param w:
        :param gsyn_exc:
        :param gsyn_inh:
        """
        raise SpynnakerException(
            "This neuron model is currently not supported by the tool chain")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Sequence, Optional, TYPE_CHECKING

import numpy
from numpy import uint8
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation


class ArrayConnector(AbstractConnector, AbstractGenerateConnectorOnHost):
    """
    Make connections using an array of integers based on the IDs
    of the neurons in the pre- and post-populations.
    """

    __slots__ = (
        "__array",
        "__array_dims",
        "__n_total_connections")

    def __init__(self, array: NDArray[uint8], safe: bool = True,
                 callback: None = None, verbose: bool = False):
        """
        :param array:
            An explicit Boolean matrix that specifies the connections
            between the pre- and post-populations
            (see PyNN documentation). Must be 2D in practice.
        :param safe:
            Whether to check that weights and delays have valid values.
            If False, this check is skipped.
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        """
        super().__init__(safe, callback, verbose)
        self.__array = array
        # we can get the total number of connections straight away
        # from the Boolean matrix
        n_total_connections = 0
        # array shape
        dims = array.shape
        for i in range(dims[0]):
            for j in range(dims[1]):
                if array[i, j] == 1:
                    n_total_connections += 1

        self.__n_total_connections = n_total_connections
        self.__array_dims = dims

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays, len(self.__array), synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays, len(self.__array), synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        # Break the array into n_post_atoms units
        split_positions = numpy.arange(
            0, synapse_info.n_post_neurons, n_post_atoms)
        split_array = numpy.array_split(self.__array, split_positions)

        # Sum the 1s in each split row
        sum_rows = [numpy.sum(s, axis=1) for s in split_array]

        # Find the maximum of the rows
        max_connections_row = max([x for y in sum_rows for x in y])

        if min_delay is None or max_delay is None:
            return max_connections_row

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays, self.__n_total_connections,
            max_connections_row, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        # Max number per column is required
        max_connections_col = 0
        for j in range(self.__array_dims[1]):
            n_connections_col = 0
            for i in range(self.__array_dims[0]):
                if self.__array[i, j] == 1:
                    n_connections_col = 0

            if n_connections_col > max_connections_col:
                max_connections_col = n_connections_col

        return max_connections_col

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_weight_maximum(
            synapse_info.weights, self.__n_total_connections, synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        pre_neurons = []
        post_neurons = []
        n_connections = 0
        for i in range(synapse_info.n_pre_neurons):
            for j in post_vertex_slice.get_raster_ids():
                if self.__array[i, j] == 1:
                    pre_neurons.append(i)
                    post_neurons.append(j)
                    n_connections += 1

        # Feed the arrays calculated above into the block structure
        block = numpy.zeros(
            n_connections, dtype=AbstractConnector.NUMPY_SYNAPSES_DTYPE)
        block["source"] = synapse_info.pre_vertex.get_key_ordered_indices(
            numpy.array(pre_neurons))
        block["target"] = post_vertex_slice.get_relative_indices(
            numpy.array(post_neurons))
        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return f"ArrayConnector({self.__array})"



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.data import SpynnakerDataView

from .abstract_input_type import AbstractInputType

TIME_STEP = "time_step"


class InputTypeDelta(AbstractInputType):
    """
    The delta input type.
    """
    __slots__ = ()

    def __init__(self) -> None:
        super().__init__(
            # scale_factor, calculated from timestep
            [Struct([(DataType.S1615, TIME_STEP)])],
            dict())

    @overrides(AbstractInputType.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[TIME_STEP] = SpynnakerDataView.get_simulation_time_step_ms()

    @overrides(AbstractInputType.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        pass

    @overrides(AbstractInputType.get_global_weight_scale)
    def get_global_weight_scale(self) -> float:
        return 1.0



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional

from spinn_utilities.overrides import overrides

from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotSpeaker)
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol

from .push_bot_device import PushBotEthernetDevice


class PushBotEthernetSpeakerDevice(
        PushBotEthernetDevice, AbstractSendMeMulticastCommandsVertex):
    """
    The Speaker of a PushBot.
    """

    def __init__(
            self, speaker: PushBotSpeaker,
            protocol: MunichIoSpiNNakerLinkProtocol,
            start_active_time: int = 0, start_total_period: int = 0,
            start_frequency: Optional[int] = 0,
            start_melody: Optional[int] = None,
            timesteps_between_send: Optional[int] = None):
        """
        :param speaker: The speaker to control
        :param protocol: The protocol instance to get commands from
        :param start_active_time: The "active time" to set at the start
        :param start_total_period: The "total period" to set at the start
        :param start_frequency: The "frequency" to set at the start
        :param start_melody: The "melody" to set at the start
        :param timesteps_between_send:
            The number of timesteps between sending commands to the device,
            or `None` to use the default
        """
        if not isinstance(speaker, PushBotSpeaker):
            raise ConfigurationException(
                "speaker parameter must be a PushBotSpeaker value")

        super().__init__(protocol, speaker, True, timesteps_between_send)

        # protocol specific data items
        self.__command_protocol = protocol
        self.__start_active_time = start_active_time
        self.__start_total_period = start_total_period
        self.__start_frequency = start_frequency
        self.__start_melody = start_melody

    @overrides(PushBotEthernetDevice.set_command_protocol)
    def set_command_protocol(
            self, command_protocol: MunichIoSpiNNakerLinkProtocol) -> None:
        self.__command_protocol = command_protocol

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        # add mode command if not done already
        if not self.protocol.sent_mode_command():
            yield self.protocol.set_mode()

        # device specific commands
        yield self.__command_protocol.push_bot_speaker_config_total_period(
            total_period=self.__start_total_period)
        yield self.__command_protocol.push_bot_speaker_config_active_time(
            active_time=self.__start_active_time)
        if self.__start_frequency is not None:
            yield self.__command_protocol.push_bot_speaker_set_tone(
                frequency=self.__start_frequency)
        if self.__start_melody is not None:
            yield self.__command_protocol.push_bot_speaker_set_melody(
                melody=self.__start_melody)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        yield self.__command_protocol.push_bot_speaker_config_total_period(0)
        yield self.__command_protocol.push_bot_speaker_config_active_time(0)
        yield self.__command_protocol.push_bot_speaker_set_tone(0)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import math
from typing import Optional, Sequence, TYPE_CHECKING

import numpy
from numpy import (
    arccos, arcsin, arctan, arctan2, ceil, cos, cosh, exp, fabs, floor, fmod,
    hypot, ldexp, log, log10, modf, power, sin, sinh, sqrt, tan, tanh, maximum,
    minimum, e, pi)
from numpy.typing import NDArray

from pyNN.random import NumpyRNG

from spinn_utilities.overrides import overrides
from spinn_utilities.safe_eval import SafeEval

from pacman.model.graphs.common import Slice

from spynnaker.pyNN.utilities import utility_calls

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation

# support for arbitrary expression for the indices
_index_expr_context = SafeEval(math, numpy, arccos, arcsin, arctan, arctan2,
                               ceil, cos, cosh, exp, fabs, floor, fmod, hypot,
                               ldexp, log, log10, modf, power, sin, sinh, sqrt,
                               tan, tanh, maximum, minimum, e=e, pi=pi)


class IndexBasedProbabilityConnector(AbstractConnector,
                                     AbstractGenerateConnectorOnHost):
    """
    Make connections using a probability distribution which varies
    dependent upon the indices of the pre- and post-populations.
    """

    __slots = [
        "__allow_self_connections",
        "__index_expression",
        "__probs",
        "__rng"]

    def __init__(
            self, index_expression: str, allow_self_connections: bool = True,
            rng: Optional[NumpyRNG] = None,
            safe: bool = True, callback: None = None, verbose: bool = False):
        """
        :param index_expression:
            the right-hand side of a valid python expression for
            probability, involving the indices of the pre- and
            post-populations,
            that can be parsed by `eval()`, that computes a probability
            distribution;
            the indices will be given as variables ``i`` and ``j`` when the
            expression is evaluated.
        :param allow_self_connections:
            if the connector is used to connect a Population to itself, this
            flag determines whether a neuron is allowed to connect to itself,
            or only to other neurons in the Population.
        :param rng:
            Seeded random number generator, or ``None`` to make one when
            needed.
        :param safe:
            Whether to check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        """
        super().__init__(safe, callback, verbose)
        self.__rng = rng or NumpyRNG()
        self.__index_expression = index_expression
        self.__allow_self_connections = allow_self_connections
        self.__probs: Optional[NDArray] = None

    def _update_probs_from_index_expression(
            self, synapse_info: SynapseInformation) -> NDArray:
        # note: this only needs to be done once
        if self.__probs is None:
            # numpy array of probabilities using the index_expression
            self.__probs = numpy.fromfunction(
                lambda i, j: _index_expr_context.eval(
                    self.__index_expression, i=i, j=j),
                (synapse_info.n_pre_neurons, synapse_info.n_post_neurons))
        return self.__probs

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        probs = self._update_probs_from_index_expression(synapse_info)
        n_connections = utility_calls.get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            numpy.amax(probs))
        return self._get_delay_maximum(
            synapse_info.delays, n_connections, synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        probs = self._update_probs_from_index_expression(synapse_info)
        n_connections = utility_calls.get_probable_minimum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            numpy.amax(probs))
        return self._get_delay_minimum(
            synapse_info.delays, n_connections, synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        probs = self._update_probs_from_index_expression(synapse_info)
        n_connections = utility_calls.get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_post_atoms, numpy.amax(probs))

        if min_delay is None or max_delay is None:
            return int(math.ceil(n_connections))

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_connections, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        probs = self._update_probs_from_index_expression(synapse_info)
        return utility_calls.get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons, numpy.amax(probs))

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(
            self, synapse_info: SynapseInformation) -> float:
        probs = self._update_probs_from_index_expression(synapse_info)
        n_connections = utility_calls.get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            numpy.amax(probs))
        return self._get_weight_maximum(
            synapse_info.weights, n_connections, synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        # setup probabilities here
        probs = self._update_probs_from_index_expression(synapse_info)

        probs = probs[:, post_vertex_slice.get_raster_ids()].reshape(-1)

        n_items = synapse_info.n_pre_neurons * post_vertex_slice.n_atoms
        items = self.__rng.next(n_items)

        # If self connections are not allowed, remove the possibility of self
        # connections by setting the probability to a value of infinity
        no_self = (
            not self.__allow_self_connections and
            synapse_info.pre_population == synapse_info.post_population)
        if no_self:
            items[0:n_items:post_vertex_slice.n_atoms + 1] = numpy.inf

        present = items < probs
        ids = numpy.where(present)[0]
        n_connections = numpy.sum(present)

        block = numpy.zeros(
            n_connections, dtype=AbstractConnector.NUMPY_SYNAPSES_DTYPE)
        block["source"] = synapse_info.pre_vertex.get_key_ordered_indices(
            ids // post_vertex_slice.n_atoms)
        block["target"] = ids % post_vertex_slice.n_atoms
        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return f"IndexBasedProbabilityConnector({self.__index_expression})"

    @property
    def allow_self_connections(self) -> bool:
        """
        When the connector is used to connect a Population to itself, this
        flag determines whether a neuron is allowed to connect to itself,
        or only to other neurons in the Population.
        """
        return self.__allow_self_connections

    @allow_self_connections.setter
    def allow_self_connections(self, new_value: bool) -> None:
        self.__allow_self_connections = new_value

    @property
    def index_expression(self) -> str:
        """
        The right-hand side of a valid python expression for probability,
        involving the indices of the pre- and post-populations, that can
        be parsed by `eval()`, that computes a probability distribution.
        """
        return self.__index_expression

    @index_expression.setter
    def index_expression(self, new_value: str) -> None:
        if self.__probs is None:
            raise ValueError("connectivity matrix already fixed")
        self.__index_expression = new_value



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional
from spinn_front_end_common.utilities.connections import LiveEventConnection
from spinn_front_end_common.utilities.constants import NOTIFY_PORT

# The maximum number of 32-bit keys that will fit in a packet
_MAX_FULL_KEYS_PER_PACKET = 63
# The maximum number of 16-bit keys that will fit in a packet
_MAX_HALF_KEYS_PER_PACKET = 127


class SpynnakerLiveSpikesConnection(LiveEventConnection):
    """
    A connection for receiving and sending live spikes from and to
    SpiNNaker.
    """
    __slots__ = ()

    def __init__(self, receive_labels: Optional[Iterable[str]] = None,
                 send_labels: Optional[Iterable[str]] = None,
                 local_host: Optional[str] = None,
                 local_port: Optional[int] = NOTIFY_PORT,
                 live_packet_gather_label: str = "LiveSpikeReceiver"):
        """
        :param receive_labels:
            Labels of population from which live spikes will be received.
        :param send_labels:
            Labels of population to which live spikes will be sent
        :param local_host:
            Optional specification of the local hostname or IP address of the
            interface to listen on
        :param local_port:
            Optional specification of the local port to listen on. Must match
            the port that the toolchain will send the notification on (19999
            by default)
        """
        super().__init__(
            live_packet_gather_label, receive_labels, send_labels,
            local_host, local_port)

    def send_spike(self, label: str, neuron_id: int,
                   send_full_keys: bool = False) -> None:
        """
        Send a spike from a single neuron.

        :param label:
            The label of the population from which the spike will originate
        :param neuron_id: The ID of the neuron sending a spike
        :param send_full_keys: Determines whether to send full 32-bit
            keys, getting the key for each neuron from the database, or
            whether to send 16-bit neuron IDs directly
        """
        self.send_spikes(label, [neuron_id], send_full_keys)

    def send_spikes(
            self, label: str, neuron_ids: List[int],
            send_full_keys: bool = False) -> None:
        """
        Send a number of spikes.

        :param label:
            The label of the population from which the spikes will originate
        :param neuron_ids: array-like of neuron IDs sending spikes
        :param send_full_keys: Determines whether to send full 32-bit
            keys, getting the key for each neuron from the database, or
            whether to send 16-bit neuron IDs directly
        """
        self.send_events(label, neuron_ids, send_full_keys)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from math import ceil, log2
from typing import Dict, Optional
import logging
from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter
from pacman.model.graphs.application import Application2DSpiNNakerLinkVertex
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex
from pacman.model.routing_info.base_key_and_mask import BaseKeyAndMask
from pacman.utilities.constants import BITS_IN_KEY
from pacman.utilities.utility_calls import is_power_of_2
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = FormatAdapter(logging.getLogger(__name__))

DEFAULT_WIDTH = 304
DEFAULT_HEIGHT = 240


class ICUBRetinaDevice(Application2DSpiNNakerLinkVertex):
    """
    An ICUB retina device connected to SpiNNaker using a SpiNNakerLink.
    """

    __slots__ = (
        "__index_by_slice",
        "__base_key")

    def __init__(self, base_key: int = 0, width: int = DEFAULT_WIDTH,
                 height: int = DEFAULT_HEIGHT,
                 sub_width: int = 16, sub_height: int = 16,
                 spinnaker_link_id: int = 0,
                 board_address: Optional[str] = None):
        """
        :param base_key: The key that is common over the whole vertex
        :param width: The width of the retina in pixels
        :param height: The height of the retina in pixels
        :param sub_width:
            The width of rectangles to split the retina into for efficiency of
            sending
        :param sub_height:
            The height of rectangles to split the retina into for efficiency of
            sending
        :param spinnaker_link_id:
            The ID of the SpiNNaker link that the device is connected to
        :param board_address:
            The board to which the device is connected,
            or `None` for the first board
        """
        # Fake the width if not a power of 2, as we need this for the sake
        # of passing on to other 2D vertices.  This will map the pixels onto
        # an image that is bigger than the real image size so will have blank
        # sections that are never used.  This could result in more cores than
        # necessary being used by downstream populations, but this is hard to
        # avoid!
        if not is_power_of_2(width):
            if width == DEFAULT_WIDTH:
                width = 2 ** int(ceil(log2(width)))
                logger.warning(
                    "The width of the ICUB retina has been rounded up from {}"
                    " to {}.  This is to ensure that the coordinates are"
                    " mapped correcting in following Populations.  Note that"
                    " this might mean that these Populations then also need to"
                    " be bigger to match this (and you might get an error"
                    " message for some e.g. Convolutional populations)",
                    DEFAULT_WIDTH, width)
            else:
                raise ConfigurationException(
                    "The width of the ICUB retina must be a power of 2.  If"
                    " the real retina size is less than this, please round it"
                    " up. This will ensure that following Populations can"
                    " decode the spikes correctly.  Note that you will also"
                    " have to make the sizes of the following Populations"
                    " bigger to match!")

        # Call the super
        super().__init__(
            width, height, sub_width, sub_height, spinnaker_link_id,
            board_address, incoming=True, outgoing=True)

        # A dictionary to get vertex index from FPGA and slice
        self.__index_by_slice: Dict[Slice, int] = dict()
        self.__base_key = base_key

    @overrides(Application2DSpiNNakerLinkVertex.get_incoming_slice)
    def get_incoming_slice(self, index: int) -> Slice:
        vertex_slice = super().get_incoming_slice(index)
        self.__index_by_slice[vertex_slice] = index
        return vertex_slice

    @overrides(Application2DSpiNNakerLinkVertex.get_machine_fixed_key_and_mask)
    def get_machine_fixed_key_and_mask(
            self, machine_vertex: MachineVertex,
            partition_id: str) -> BaseKeyAndMask:
        vertex_slice = machine_vertex.vertex_slice
        index = self.__index_by_slice[vertex_slice]
        return self._get_key_and_mask(self.__base_key, index)

    @overrides(Application2DSpiNNakerLinkVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(self, partition_id: str) -> BaseKeyAndMask:
        n_key_bits = BITS_IN_KEY - self._key_shift
        key_mask = ((1 << n_key_bits) - 1) << self._key_shift
        return BaseKeyAndMask(self.__base_key << self._key_shift, key_mask)

    @property
    def _source_x_mask(self) -> int:
        return ((1 << self._x_bits) - 1) << 1

    @property
    def _source_x_shift(self) -> int:
        return 1

    @property
    def _source_y_mask(self) -> int:
        return ((1 << self._y_bits) - 1) << 12

    @property
    def _source_y_shift(self) -> int:
        return 12

    @property
    def _key_shift(self) -> int:
        return 20



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.progress_bar import ProgressBar
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neural_projections import ProjectionApplicationEdge


def finish_connection_holders() -> None:
    """
    Finishes the connection holders after data has been generated within
    them, allowing any waiting callbacks to be called.
    """
    edges = SpynnakerDataView.get_edges()
    progress = ProgressBar(len(edges), "Finalising Retrieved Connections")
    for edge in progress.over(edges):
        if isinstance(edge, ProjectionApplicationEdge):
            for synapse_info in edge.synapse_information:
                synapse_info.finish_connection_holders()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Dict, Iterable, Optional, Tuple
from spinn_utilities.overrides import overrides
from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.connections.live_event_connection \
    import (
        LiveEventConnection, _Callback, _InitCallback, _RcvCallback,
        _RcvTimeCallback)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.constants import NOTIFY_PORT


class SpynnakerPoissonControlConnection(LiveEventConnection):
    """
    A connection used to control a Poisson-distributed random event source's
    firing rate at runtime.
    """
    __slots__ = (
        "__control_label_extension",
        "__control_label_to_label",
        "__label_to_control_label")

    def __init__(
            self, poisson_labels: Optional[Iterable[str]] = None,
            local_host: Optional[str] = None,
            local_port: Optional[int] = NOTIFY_PORT,
            control_label_extension: str = "_control"):
        """
        :param poisson_labels:
            Labels of Poisson populations to be controlled
        :param local_host: Optional specification of the local hostname or
            IP address of the interface to listen on
        :param local_port:
            Optional specification of the local port to listen on. Must match
            the port that the toolchain will send the notification on (19999
            by default)
        :param control_label_extension:
            The extra name added to the label of each Poisson source
        """
        self.__control_label_extension = control_label_extension

        control_labels: Optional[Iterable[str]] = None
        self.__control_label_to_label: Dict[str, str] = dict()
        self.__label_to_control_label: Dict[str, str] = dict()
        if poisson_labels is not None:
            control_labels = [
                self.__convert_to_control_label(label)
                for label in poisson_labels
            ]
            self.__control_label_to_label.update(
                {control: label
                 for control, label in zip(control_labels, poisson_labels)})
            self.__label_to_control_label.update(
                {label: control
                 for label, control in zip(poisson_labels, control_labels)})

        super().__init__(
            live_packet_gather_label=None, send_labels=control_labels,
            local_host=local_host, local_port=local_port)

    def add_poisson_label(self, label: str) -> None:
        """
        Adds a send label.

        :param label: The label of the Poisson source population.
        """
        control = self.__convert_to_control_label(label)
        self.__control_label_to_label[control] = label
        self.__label_to_control_label[label] = control
        self.add_send_label(control)

    def __convert_to_control_label(self, label: str) -> str:
        return f"{label}{self.__control_label_extension}"

    def __control_label(self, label: str) -> str:
        # Try to get a control label, but if not just use the label
        return self.__label_to_control_label.get(label, label)

    def __label(self, control_label: str) -> str:
        # Try to get a label, but if not just use the control label
        return self.__control_label_to_label.get(control_label, control_label)

    def __callback_wrapper(
            self, callback: _Callback, label: str,
            connection: LiveEventConnection) -> None:
        callback(self.__label(label), connection)

    def __init_callback_wrapper(
            self, callback: _InitCallback,
            label: str, vertex_size: int, run_time_ms: float,
            machine_timestep_ms: float) -> None:
        callback(self.__label(label), vertex_size, run_time_ms,
                 machine_timestep_ms)

    @overrides(LiveEventConnection.add_start_callback)
    def add_start_callback(
            self, label: str, start_callback: _Callback) -> None:
        super().add_start_callback(
            self.__control_label(label), functools.partial(
                self.__callback_wrapper, start_callback))

    @overrides(LiveEventConnection.add_start_resume_callback)
    def add_start_resume_callback(
            self, label: str, start_resume_callback: _Callback) -> None:
        super().add_start_resume_callback(
            self.__control_label(label), functools.partial(
                self.__callback_wrapper, start_resume_callback))

    @overrides(LiveEventConnection.add_init_callback)
    def add_init_callback(
            self, label: str, init_callback: _InitCallback) -> None:
        super().add_init_callback(
            self.__control_label(label), functools.partial(
                self.__init_callback_wrapper, init_callback))

    @overrides(LiveEventConnection.add_receive_callback)
    def add_receive_callback(
            self, label: str, live_event_callback: _RcvTimeCallback,
            translate_key: bool = True) -> None:
        raise ConfigurationException(
            "SpynnakerPoissonControlPopulation can't receive data")

    @overrides(LiveEventConnection.add_receive_no_time_callback)
    def add_receive_no_time_callback(
            self, label: str, live_event_callback: _RcvCallback,
            translate_key: bool = True) -> None:
        raise ConfigurationException(
            "SpynnakerPoissonControlPopulation can't receive data")

    @overrides(LiveEventConnection.add_pause_stop_callback)
    def add_pause_stop_callback(
            self, label: str, pause_stop_callback: _Callback) -> None:
        super().add_pause_stop_callback(
            self.__control_label(label), functools.partial(
                self.__callback_wrapper, pause_stop_callback))

    def set_rate(self, label: str, neuron_id: int, rate: float) -> None:
        """
        Set the rate of a Poisson neuron within a Poisson source.

        :param label: The label of the Population to set the rates of
        :param neuron_id: The neuron ID to set the rate of
        :param rate: The rate to set in Hz
        """
        self.set_rates(label, [(neuron_id, rate)])

    def set_rates(self, label: str,
                  neuron_id_rates: Iterable[Tuple[int, float]]) -> None:
        """
        Set the rates of multiple Poisson neurons within a Poisson source.

        :param label: The label of the Population to set the rates of
        :param neuron_id_rates:
            A list of tuples of (neuron ID, rate) to be set
        """
        control = self.__control_label(label)
        datatype = DataType.U1616
        atom_ids_and_payloads = [(nid, datatype.encode_as_int(rate))
                                 for nid, rate in neuron_id_rates]
        self.send_events_with_payloads(control, atom_ids_and_payloads)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Mapping
from spinn_utilities.overrides import overrides
from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SpynnakerException
from .abstract_current_source import (
    AbstractCurrentSource, CurrentSourceIDs, CurrentParameter)


class DCSource(AbstractCurrentSource):
    """
    Current source with amplitude turned on at "start" and off at "stop".
    """
    __slots__ = (
        "__amplitude",
        "__start",
        "__stop",
        "__parameters",
        "__parameter_types")

    def __init__(self, amplitude: float = 0.0, start: float = 0.0,
                 stop: float = 0.0) -> None:
        """
        :param amplitude:
        :param start:
        :param stop:
        """
        # There's probably no need to actually store these as you can't
        # access them directly in pynn anyway
        self.__amplitude = amplitude
        self.__start = start
        self.__stop = stop

        self.__parameter_types = {
            'amplitude': DataType.S1615,
            # Anything associated with timing needs to be an integer on
            # the machine
            'start': DataType.UINT32,
            'stop': DataType.UINT32}

        time_convert_ms = SpynnakerDataView.get_simulation_time_step_per_ms()
        self.__parameters: Dict[str, CurrentParameter] = {
            'amplitude': self.__amplitude,
            # Convert to integers i.e. timesteps
            'start': int(self.__start * time_convert_ms),
            'stop': int(self.__stop * time_convert_ms)}

        super().__init__()

    @overrides(AbstractCurrentSource.set_parameters)
    def set_parameters(self, **parameters: CurrentParameter) -> None:
        for key, value in parameters.items():
            if key not in self.__parameters:
                # throw an exception
                raise SpynnakerException(f"{key} is not a parameter of {self}")
            self.__parameters[key] = value

        # Parameters have been set, so if multi-run then it will have been
        # injected already; if not then it can just be ignored
        if self.app_vertex is not None:
            for m_vertex in self.app_vertex.machine_vertices:
                m_vertex.set_reload_required(True)

    @property
    @overrides(AbstractCurrentSource.parameters)
    def parameters(self) -> Mapping[str, CurrentParameter]:
        return self.__parameters

    @property
    @overrides(AbstractCurrentSource.parameter_types)
    def parameter_types(self) -> Mapping[str, DataType]:
        return self.__parameter_types

    @property
    @overrides(AbstractCurrentSource.current_source_id)
    def current_source_id(self) -> int:
        return CurrentSourceIDs.DC_SOURCE.value

    @overrides(AbstractCurrentSource.get_sdram_usage_in_bytes)
    def get_sdram_usage_in_bytes(self) -> int:
        return len(self.__parameters) * BYTES_PER_WORD



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
import math
from typing import Optional, Sequence, TYPE_CHECKING

import numpy
from numpy.typing import NDArray

from pyNN.random import NumpyRNG

from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter

from pacman.model.graphs.common import Slice

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.utilities.utility_calls import (
    get_probable_maximum_selected, get_probable_minimum_selected, check_rng)
from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)

logger = FormatAdapter(logging.getLogger(__name__))

N_GEN_PARAMS = 6


class FixedProbabilityConnector(AbstractGenerateConnectorOnMachine,
                                AbstractGenerateConnectorOnHost):
    """
    For each pair of pre-post cells, the connection probability is constant.
    """

    __slots__ = (
        "__allow_self_connections",
        "_p_connect",
        "__rng")

    def __init__(
            self, p_connect: float, allow_self_connections: bool = True,
            safe: bool = True, verbose: bool = False,
            rng: Optional[NumpyRNG] = None, callback: None = None):
        """
        :param p_connect:
            a value between zero and one. Each potential connection is created
            with this probability.
        :param allow_self_connections:
            if the connector is used to connect a Population to itself, this
            flag determines whether a neuron is allowed to connect to itself,
            or only to other neurons in the Population.
        :param safe:
            If ``True``, check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param rng:
            Seeded random number generator, or `None` to make one when needed
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        # Support 1.0 by using maximum U032.  Warn the user because this isn't
        # *quite* the same
        if p_connect == 1.0:
            p_connect = float(DataType.U032.max)
            logger.warning(
                "Probability of 1.0 in the FixedProbabilityConnector will use "
                "{} instead.  If this is a problem, use the AllToAllConnector "
                "instead.", p_connect)
        if not 0.0 <= p_connect < 1.0:
            raise ConfigurationException(
                "The probability must be >= 0 and < 1")
        super().__init__(safe, callback, verbose)
        self._p_connect = p_connect
        self.__allow_self_connections = allow_self_connections
        self.__rng = rng

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        n_connections = get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            self._p_connect)
        return self._get_delay_maximum(
            synapse_info.delays, n_connections, synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        n_connections = get_probable_minimum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            self._p_connect)
        return self._get_delay_minimum(
            synapse_info.delays, n_connections, synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        n_connections = get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_post_atoms, self._p_connect, chance=1.0/10000.0)

        if min_delay is None or max_delay is None:
            return int(math.ceil(n_connections))

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_connections, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        n_connections = get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons, self._p_connect,
            chance=1.0/10000.0)
        return n_connections

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        n_connections = get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            self._p_connect)
        return self._get_weight_maximum(
            synapse_info.weights, n_connections, synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        rng = self.__rng or NumpyRNG()
        n_items = synapse_info.n_pre_neurons * post_vertex_slice.n_atoms
        items = rng.next(n_items)

        # If self connections are not allowed, remove possibility the self
        # connections by setting them to a value of infinity
        no_self = (
            not self.__allow_self_connections and
            synapse_info.pre_population == synapse_info.post_population)
        if no_self:
            items[0:n_items:post_vertex_slice.n_atoms + 1] = numpy.inf

        present = items < self._p_connect
        ids = numpy.where(present)[0]
        n_connections = numpy.sum(present)

        block = numpy.zeros(n_connections, dtype=self.NUMPY_SYNAPSES_DTYPE)
        block["source"] = ids // post_vertex_slice.n_atoms
        block["target"] = (
            (ids % post_vertex_slice.n_atoms) + post_vertex_slice.lo_atom)
        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return f"FixedProbabilityConnector({self._p_connect})"

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.FIXED_PROBABILITY_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray:
        allow_self = (
            self.__allow_self_connections or
            synapse_info.pre_population != synapse_info.post_population)
        return numpy.array([
            int(allow_self),
            DataType.U032.encode_as_int(self._p_connect)], dtype="uint32")

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        return 2 * BYTES_PER_WORD

    @property
    def p_connect(self) -> float:
        """
        Probability for each potential connection.

        A value between zero and one. (inclusive)
        """
        return self._p_connect

    @p_connect.setter
    def p_connect(self, new_value: float) -> None:
        if not 0.0 <= new_value <= 1.0:
            raise ConfigurationException(
                "The probability must be between 0 and 1 (inclusive)")
        self._p_connect = new_value

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        if self.generate_on_machine(synapse_info):
            check_rng(self.__rng, "FixedProbabilityConnector")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Mapping, Optional, Sequence
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.ranged import RangeDictionary
from spinn_front_end_common.interface.ds import DataType
from spynnaker.pyNN.utilities.struct import Struct


class AbstractNeuronImpl(object, metaclass=AbstractBase):
    """
    An abstraction of a whole neuron model including all parts.
    """

    __slots__ = ()

    @property
    @abstractmethod
    def model_name(self) -> str:
        """
        The name of the model.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def binary_name(self) -> str:
        """
        The name of the binary executable of this implementation.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def structs(self) -> Sequence[Struct]:
        """
        A list of structures used by the implementation.
        """
        raise NotImplementedError

    @abstractmethod
    def get_global_weight_scale(self) -> float:
        """
        :returns: The weight scaling required by this model.
        """
        raise NotImplementedError

    @abstractmethod
    def get_n_synapse_types(self) -> int:
        """
        :returns: The number of synapse types supported by the model.
        """
        raise NotImplementedError

    @abstractmethod
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        """
        :param target: The name of the synapse
        :returns: The ID of a synapse given the name.
        """
        raise NotImplementedError

    @abstractmethod
    def get_synapse_targets(self) -> Sequence[str]:
        """
        :returns: The target names of the synapse type.
        """
        raise NotImplementedError

    @abstractmethod
    def get_recordable_variables(self) -> Sequence[str]:
        """
        :returns:
           The names of the variables that can be recorded in this model.
        """
        raise NotImplementedError

    @abstractmethod
    def get_recordable_units(self, variable: str) -> str:
        """
        Get the units of the given variable that can be recorded.

        :param variable: The name of the variable
        :returns: The unit or this variable. For example 'mV' or 'uS'.
           Will be an empty string for things like spikes and probability
        """
        raise NotImplementedError

    @abstractmethod
    def get_recordable_data_types(self) -> Mapping[str, DataType]:
        """
        Get the data type of the variables that can be recorded.

        :return: dictionary of name of variable to data type of variable
        """
        raise NotImplementedError

    @abstractmethod
    def is_recordable(self, variable: str) -> bool:
        """
        Determine if the given variable can be recorded.

        :param variable: The name of the variable
        :returns: True if the variable is recorded, False otherwise
        """
        raise NotImplementedError

    @abstractmethod
    def get_recordable_variable_index(self, variable: str) -> int:
        """
        :param variable: The name of the variable
        :returns: The index of the variable in the list of variables that can
            be recorded.
        """
        raise NotImplementedError

    @abstractmethod
    def add_parameters(self, parameters: RangeDictionary) -> None:
        """
        Add the initial values of the parameters to the parameter holder.

        :param parameters: A holder of the parameters
        """
        raise NotImplementedError

    @abstractmethod
    def add_state_variables(self, state_variables: RangeDictionary) -> None:
        """
        Add the initial values of the state variables to the state
        variables holder.

        :param state_variables: A holder of the state variables
        """
        raise NotImplementedError

    @abstractmethod
    def get_units(self, variable: str) -> str:
        """
        :param variable: The name of the variable
        :returns: The units of the given variable.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def is_conductance_based(self) -> bool:
        """
        Whether the model uses conductance.
        """
        raise NotImplementedError
